'''

This module contains various classes that can be used for specifying different
types of samplers. These different samplers implement basic sampling 
techniques including Full Factorial sampling, Latin Hypercube sampling, and
Monte Carlo sampling.

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)
from ema_workbench.em_framework.parameters import CategoricalParameter

try:
    from future_builtins import zip
except ImportError:
    try:
        from itertools import izip as zip # < 2.5 or 3.x
    except ImportError:
        pass

import abc
import functools
import itertools
import numpy as np 
import operator
import scipy.stats as stats

from . import util
from .parameters import (IntegerParameter, Policy, Scenario)

# Created on 16 aug. 2011
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['AbstractSampler',
           'LHSSampler',
           'MonteCarloSampler',
           'FullFactorialSampler',
           'PartialFactorialSampler',
           'sample_levers',
           'sample_uncertainties',
           'determine_parameters']


class AbstractSampler(object):
    '''
    Abstract base class from which different samplers can be derived. 
    
    In the simplest cases, only the sample method needs to be overwritten. 
    generate_designs` is the only method called by the ensemble class. The 
    other methods are used internally to generate the designs. 
    
    '''
    __metaaclass__ = abc.ABCMeta
    
    # types of distributions known by the sampler.
    # by default it knows the `uniform continuous <http://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.uniform.html>`_
    # distribution for sampling floats, and the `uniform discrete <http://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.randint.html#scipy.stats.randint>`_
    # distribution for sampling integers. 
    distributions = {"uniform" : stats.uniform,
                     "integer" : stats.randint
                              }
    
    def __init__(self):
        super(AbstractSampler, self).__init__()

    @abc.abstractmethod
    def sample(self, distribution, params, size):
        '''
        method for sampling a number of samples from a particular distribution.
        The various samplers differ with respect to their implementation of 
        this method. 
        
        Parameters
        ----------
        distribution : {'uniform', 'integer'} 
                       the distribution to sample from
        params : tuple
                 the parameters specifying the distribution
        size : int 
               the number of samples to generate
        
        Returns
        -------
        numpy array
            the samples for the distribution and specified parameters
        
        '''
    
    def generate_samples(self, parameters, size):
        '''
        The main method of :class: `~sampler.Sampler` and its 
        children. This will call the sample method for each of the 
        parameters and return the resulting designs. 
        
        Parameters
        ----------
        parameters : collection
                     a collection of :class:`~parameters.Parameterparamertainty` 
                     and :class:`~parameters.Categoricalparamertainty`
                     instances.
        size : int
               the number of samples to generate.
        
        
        Returns
        -------
        dict
            dict with the paramertainty.name as key, and the sample as value
        
        '''
        return {param.name: self.sample(param.dist, param.params, size) for 
                param in parameters}

    def generate_designs(self,  parameters, nr_samples):
        '''external interface to sampler. Returns the computational experiments
        over the specified parameters, for the given number of samples for each
        parameter.
        
        Parameters
        ----------
        parameters : list 
                        a list of parameters for which to generate the
                        experimental designs
        nr_samples : int
                     the number of samples to draw for each parameter
        
        
        Returns
        -------
        generator
            a generator object that yields the designs resulting from
            combining the parameters
        int
            the number of experimental designs
        
        '''
        parameters = sorted(parameters, key=operator.attrgetter('name'))
        sampled_parameters = self.generate_samples(parameters, nr_samples)
        designs = zip(*[sampled_parameters[u.name] for u in parameters]) 
        designs = DefaultDesigns(designs, parameters, nr_samples)
        
        return designs


class LHSSampler(AbstractSampler):
    """
    generates a Latin Hypercube sample for each of the parameters
    in case of categorical parameters, it handles the transform as well
    """
    
    def __init__(self):
        super(LHSSampler, self).__init__()
    
    def sample(self, distribution, params, size):
        '''
        generate a Latin Hypercube Sample.
        
        Parameters
        ----------
        distribution : scipy distribution
                       the distribution to sample from
        params : tuple
                 the parameters specifying the distribution
        size : int
               the number of samples to generate
    
        Returns
        -------
        dict 
            with the paramertainty.name as key, and the sample as value
    
        '''
        
        return self._lhs(self.distributions[distribution], params, size)

    def _lhs(self, dist, parms, siz):
        '''
        Latin Hypercube sampling of any distribution.
    
        Parameters
        ----------
        dist : random number generator from `scipy.stats <http://docs.scipy.org/doc/scipy/reference/stats.html>`_
        parms : tuple
                tuple of parameters as required for dist.
        siz : int 
              number of samples
    
        '''
        perc = np.arange(0,1.,1./siz)
        np.random.shuffle(perc)
        smp = stats.uniform(perc,1./siz).rvs() 
        v = dist(*parms).ppf(smp)
        return v
    
        
class MonteCarloSampler(AbstractSampler):
    """
    generates a Monte Carlo sample for each of the parameters. 
    
    
    In case of a CategoricalParameter it also handles the transformation from 
    integers back to categories
    """
    
    def __init__(self):
        super(MonteCarloSampler, self).__init__()
        
    def sample(self, distribution, params, size):
        '''
        generate a Monte Carlo Sample.
        
        Parameters
        ----------
        distribution : scipy distribution
                       the distribution to sample from
        params : 2-tuple of floats
                 the parameters specifying the distribution
        size : int
               the number of samples to generate
        
        Returns
        -------
        dict 
            with the paramertainty.name as key, and the sample as value
               
        '''
        
        return self.distributions[distribution](*params).rvs(size)


class FullFactorialSampler(AbstractSampler):     
    '''
    generates a full factorial sample.
    
    If the parameter is non categorical, the resolution is set the 
    number of samples. If the parameter is categorical, the specified value 
    for samples will be ignored and each category will be used instead. 
    
    '''
    
    def __init__(self):
        super(FullFactorialSampler, self).__init__()
       
    def generate_samples(self,  parameters, size):
        '''
        The main method of :class: `~sampler.Sampler` and its 
        children. This will call the sample method for each of the 
        parameters and return the resulting samples 
        
        Parameters
        ----------
        parameters : collection
                        a collection of :class:`~parameters.Parameter`
                        instances
        size : int
                the number of samples to generate.
        
        Returns
        -------
        dict 
            with the paramertainty.name as key, and the sample as value
        '''
        samples = {}
        for param in parameters:
            cats = param.resolution
            if not cats:
                cats = np.linspace(param.lower_bound, 
                                   param.upper_bound, 
                                   size)
                if isinstance(param, IntegerParameter):
                    cats = np.round(cats, 0)
                    cats = set(cats)
                    cats = (int(entry) for entry in cats)
                    cats = sorted(cats)
            samples[param.name] = cats
        
        return samples
       
    def generate_designs(self,  parameters, nr_samples):
        '''
        This method provides an alternative implementation to the default 
        implementation provided by :class:`~sampler.Sampler`. This
        version returns a full factorial design across the parameters. 
        
        Parameters
        ----------
        parameters : list 
                        a list of parameters for which to generate the
                        experimental designs
        nr_samples : int
                     the number of intervals to use on each
                     Parameter. Categorical parameters always
                     return all their categories
        
        Returns
        -------
        generator
            a generator object that yields the designs resulting from
            combining the parameters
        int
            the number of experimental designs
        
        '''
        parameters = sorted(parameters, key=operator.attrgetter('name'))
        
        samples = self.generate_samples(parameters, nr_samples)
        zipped_samples = itertools.product(*[samples[u.name] for u in parameters])
        
        n_designs = self.determine_nr_of_designs(samples)
        designs = DefaultDesigns(zipped_samples, parameters, n_designs)
        
        return designs

    def determine_nr_of_designs(self, sampled_parameters):
        '''
        Helper function for determining the number of experiments that will
        be generated given the sampled parameters.
        
        Parameters
        ----------
        sampled_parameters : list 
                        a list of sampled parameters, as 
                        the values return by generate_samples
        
        Returns
        -------
        int
            the total number of experimental design
        '''
        nr_designs = 1
        for value in sampled_parameters.values():
            nr_designs *= len(value)
        return nr_designs


class PartialFactorialSampler(AbstractSampler):
    """
    generates a partial factorial design over the parameters. Any parameter 
    where factorial is true will be included in a factorial design, while the 
    remainder will be sampled using LHS or MC sampling.
    
    Parameters
    ----------
    sampling: {PartialFactorialSampler.LHS, PartialFactorialSampler.MC}, optional
              the desired sampling for the non factorial parameters.
    
    Raises
    ------
    ValueError
        if sampling is not either LHS or MC
    
    """
    
    LHS = 'LHS'
    MC = 'MC'
    
    def __init__(self, sampling='LHS'):
        super(PartialFactorialSampler, self).__init__()
        
        if sampling==PartialFactorialSampler.LHS:
            self.sampler = LHSSampler()
        elif sampling==PartialFactorialSampler.MC:
            self.sampler = MonteCarloSampler()
        else:
            raise ValueError(('invalid value for sampling type, should be LHS ' 
                              'or MC'))
        self.ff = FullFactorialSampler()
    
    def _sort_parameters(self, parameters):
        '''sort parameters into full factorial and other
        
        Parameters
        ----------
        parameters : list of parameters
        
        '''
        ff_params = []
        other_params = []
        
        for param in parameters:
            if param.pff:
                ff_params.append(param)
            else:
                other_params.append(param)
                
        return ff_params, other_params
        
    
    def generate_designs(self, parameters, nr_samples):
        '''external interface to sampler. Returns the computational experiments
        over the specified parameters, for the given number of samples for each
        parameter.
        
        Parameters
        ----------
        parameters : list 
                        a list of parameters for which to generate the
                        experimental designs
        nr_samples : int
                     the number of samples to draw for each parameter
        
        Returns
        -------
        generator
            a generator object that yields the designs resulting from
            combining the parameters
        int
            the number of experimental designs
        
        '''
        
        ff_params, other_params = self._sort_parameters(parameters)
        
        # generate a design over the factorials
        # TODO update ff to use resolution if present
        ff_designs = self.ff.generate_designs(ff_params, nr_samples)
        
        # generate a design over the remainder
        # for each factorial, run the MC design
        other_designs = self.sampler.generate_designs(other_params, 
                                                              nr_samples)
        
        nr_designs = other_designs.n * ff_designs.n
        
        designs = PartialFactorialDesigns(ff_designs, other_designs, 
                                          parameters, nr_designs)
        
        return designs


    
def determine_parameters(models, attribute, union=True):
    '''determine the parameters over which to sample
     
    Parameters
    ----------
    models : a collection of AbstractModel instances
    attribute : {'uncertainties', 'levers'}
    union : bool, optional
            in case of multiple models, sample over the union of
            levers, or over the intersection of the levers
    sampler : Sampler instance, optional
     
    Returns
    -------
    collection of Parameter instances
     
    '''
    return util.determine_objects(models, attribute, union=union)

def sample_levers(models, n_samples, union=True, sampler=LHSSampler(),
                  name=util.representation):
    '''generate policies by sampling over the levers
    
    Parameters
    ----------
    models : a collection of AbstractModel instances
    n_samples : int
    union : bool, optional
            in case of multiple models, sample over the union of
            levers, or over the intersection of the levers
    sampler : Sampler instance, optional
    name : callable, optional
           a callable to generate a name given the sampled values
          for each lever
    
    Returns
    -------
    generator yielding Policy instances
    
    '''
    levers = determine_parameters(models, 'levers', union=union)
    samples = sampler.generate_designs(levers, n_samples)
    
    partial_policy = functools.partial(Policy, name=name)
    samples.kind = partial_policy
    
    return samples

def sample_uncertainties(models, n_samples, union=True, sampler=LHSSampler()):
    '''generate scenarios by sampling over the uncertainties
    
    Parameters
    ----------
    models : a collection of AbstractModel instances
    n_samples : int
    union : bool, optional
            in case of multiple models, sample over the union of
            uncertainties, or over the intersection of the uncertianties
    sampler : Sampler instance, optional
    
    Returns
    -------
    generator 
        yielding Scenario instances
    collection
        the collection of parameters over which to sample
    n_samples
        the number of scenarios (!= n_samples in case off FF sampling)
    
    
    '''
    uncertainties = determine_parameters(models, 'uncertainties', union=union)
    samples = sampler.generate_designs(uncertainties, n_samples)
    samples.kind = Scenario
    
    return samples

def from_experiments(models, experiments):
    '''generate scenarios from an existing experiments recarray
    
    Parameters
    ----------
    models : collection of AbstractModel instances
    experiments : recarray
    
    Returns
    -------
     generator 
        yielding Scenario instances
    collection
        the collection of parameters over which to sample
    n_samples
        the number of scenarios (!= n_samples in case off FF sampling)   
    
    '''
    policy_names = np.unique(experiments['policy'])
    model_names = np.unique(experiments['model'])
    
    # we sample ff over models and policies so we need to ensure
    # we only get the experiments for a single model policy combination
    logical = (experiments['model'] == model_names[0]) & \
              (experiments['policy'] == policy_names[0]) 
    
    experiments = experiments[logical]
    experiments = np.lib.recfunctions.drop_fields(experiments,  
                                                  ['model', 'policy'], 
                                                  asrecarray=True) 
    
    uncertainties = util.determine_objects(models, 'uncertainties', 
                                           union=True)
    unc_names = np.lib.recfunctions.get_names(experiments.dtype)  # @UndefinedVariable
    uncertainties = [uncertainties[unc] for unc in unc_names]

    samples = {unc:experiments[unc] for unc in unc_names}
    
    scenarios = DefaultDesigns(samples, uncertainties, experiments.shape[0])
    scenarios.kind = Scenario
    
    return scenarios 

class DefaultDesigns(object):
    '''iterable for the experimental designs'''
    
    def __init__(self, designs, parameters, n):
        self.designs = list(designs)
        self.parameters = parameters
        self.params = [p.name for p in parameters]
        self.kind = None
        self.n = n
    
    @abc.abstractmethod 
    def __iter__(self):
        '''should return iterator'''
        
        return design_generator(self.designs, self.parameters, self.kind)


class PartialFactorialDesigns(object):
    
    @property
    def kind(self):
        return self._kind
    
    @kind.setter
    def kind(self, value):
        self._kind = value
        self.ff_designs.kind = value
        self.other_designs.kind = value
    
    def __init__(self, ff_designs, other_designs, parameters, n):
        self.ff_designs = ff_designs
        self.other_designs = other_designs
        
        self.parameters = parameters
        self.params = [p.name for p in parameters]

        self._kind = None
        self.n = n
    
    def __iter__(self):
        designs =  itertools.product(self.ff_designs, self.other_designs)
        return partial_designs_generator(designs)

def partial_designs_generator(designs):
    '''generator which combines the full factorial part of the design with
    the non full factorial part into a single dict
    
    Parameters
    ----------
    designs: iterable of tuples
    
    Yields
    ------
    dict
        experimental design dict
    
    
    '''

    for design in designs:
        ff_part, other_part = design
        
        design = ff_part.copy()
        design.update(other_part)
        
        yield design

def design_generator(designs, params, kind):
    '''generator that combines the sampled parameters with their correct 
    name in order to return dicts.
    
    Parameters
    ----------
    designs : iterable of tuples
    params : iterable of str
    
    Yields
    ------
    dict
        experimental design dictionary
    
    '''
    
    for design in designs:

        design_dict = {}
        for param, value in zip(params, design):
            if isinstance(param, IntegerParameter):
                value = int(value)
            if isinstance(param, CategoricalParameter):
                # categorical parameter is an integer parameter, so
                # conversion to int is already done
                value = param.cat_for_index(value).value
            
            design_dict[param.name] = value
        
        yield kind(**design_dict)
        


'''
Samplers for working with SALib

'''
from __future__ import (unicode_literals, print_function, absolute_import,
                        division)



# Created on 12 Jan 2017
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ["SobolSampler", "MorrisSampler", "FASTSampler", 'get_SALib_problem']

from SALib.sample import saltelli, morris, fast_sampler

import operator

from .samplers import DefaultDesigns
from .parameters import CategoricalParameter, IntegerParameter


def get_SALib_problem(uncertainties):
    '''returns a dict with a problem specificatin as required by SALib'''
    
    _warning = False
    uncertainties = sorted(uncertainties, key=operator.attrgetter('name'))
    bounds = []
    
    for uncertainty in uncertainties:
        values = uncertainty.lower_bound, uncertainty.upper_bound
        bounds.append(values)

    problem = {'num_vars': len(uncertainties),
               'names': [unc.name for unc in uncertainties],
               'bounds': bounds}
    return problem

class SALibSampler(object):

    def generate_samples(self, uncertainties, size):
        '''
        The main method of :class: `~sampler.Sampler` and its 
        children. This will call the sample method for each of the 
        uncertainties and return the resulting designs. 
        
        Parameters
        ----------
        uncertainties : collection
                        a collection of :class:`~uncertainties.ParameterUncertainty` 
                        and :class:`~uncertainties.CategoricalUncertainty`
                        instances.
        size : int
               the number of samples to generate.
        
        
        Returns
        -------
        dict
            dict with the uncertainty.name as key, and the sample as value
        
        '''
        
        problem = get_SALib_problem(uncertainties)
        samples = self.sample(problem, size)
        samples = {unc.name:samples[:,i] for i, unc in enumerate(uncertainties)}
        
        # handle integer and categorical uncertainties
        for uncertainty in uncertainties:
            sample = samples[uncertainty.name]
            
            if isinstance(uncertainty, CategoricalParameter):
                sample = [uncertainty.cat_for_index(int(entry)) for entry in 
                          sample]
            elif isinstance(uncertainty, IntegerParameter):
                sample = (int(entry) for entry in sample)

            samples[uncertainty.name] = sample
        
        return samples
    
    def generate_designs(self,  parameters, nr_samples):
        '''external interface to sampler. Returns the computational experiments
        over the specified parameters, for the given number of samples for each
        parameter.
        
        Parameters
        ----------
        parameters : list 
                        a list of parameters for which to generate the
                        experimental designs
        nr_samples : int
                     the number of samples to draw for each parameter
        
        
        Returns
        -------
        generator
            a generator object that yields the designs resulting from
            combining the parameters
        int
            the number of experimental designs
        
        '''
        parameters = sorted(parameters, key=operator.attrgetter('name'))
        sampled_parameters = self.generate_samples(parameters, nr_samples)
        
        nr_designs = next(iter(sampled_parameters.values())).shape[0]

        params = sorted(sampled_parameters.keys())
        designs = zip(*[sampled_parameters[u] for u in params]) 
        designs = DefaultDesigns(designs, parameters, nr_designs)
        
        return designs

class SobolSampler(SALibSampler):
    '''Sampler generating a Sobol design using SALib
    
    Parameters
    ----------
    second_order : bool, optional
                   indicates whether second order effects should be included
    
    '''
    
    
    def __init__(self, second_order=True):
        self.second_order = second_order
        self._warning = False
        
        super(SobolSampler, self).__init__()
    
    def sample(self, problem, size):
        return saltelli.sample(problem, size, 
                                  calc_second_order=self.second_order)
    
    

class MorrisSampler(SALibSampler):
    '''Sampler generating a morris design using SALib
    
    Parameters
    ----------
    num_levels : int
        The number of grid levels
    grid_jump : int
        The grid jump size
    optimal_trajectories : int, optional
        The number of optimal trajectories to sample (between 2 and N)
    local_optimization : bool, optional
        Flag whether to use local optimization according to Ruano et al. (2012) 
        Speeds up the process tremendously for bigger N and num_levels.
        Stating this variable to be true causes the function to ignore gurobi.
    '''

    
    def __init__(self, num_levels, grid_jump, optimal_trajectories=None, 
                 local_optimization=False):
        super(MorrisSampler, self).__init__()
        self.num_levels = num_levels
        self.grid_jump = grid_jump
        self.optimal_trajectories = optimal_trajectories
        self.local_optimization = local_optimization
        

    def sample(self, problem, size):
        return morris.sample(problem, size, self.num_levels, self.grid_jump, 
                         self.optimal_trajectories, self.local_optimization)   
        
class FASTSampler(SALibSampler):
    '''Sampler generating a Fourier Amplitude Sensitivity Test (FAST) using 
    SALib
    
    Parameters
    ----------
    n : int
        The number of samples to generate
    m : int (default: 4)
        The interference parameter, i.e., the number of harmonics to sum in the
        Fourier series decomposition 
    '''

    
    def __init__(self, n, m=4):
        super(MorrisSampler, self).__init__()
        self.n = n
        self.m = m

    def sample(self, problem, size):
        return fast_sampler(self.n, self.m)  



from __future__ import (absolute_import, print_function, unicode_literals, 
                        division)

import sys
import os
import glob
from contextlib import contextmanager
import traceback

set_backend = "import matplotlib\nmatplotlib.use('Agg')\n"

@contextmanager
def redirected_output(new_stdout=None, new_stderr=None):
    save_stdout = sys.stdout
    save_stderr = sys.stderr
    if new_stdout is not None:
        sys.stdout = new_stdout
    if new_stderr is not None:
        sys.stderr = new_stderr
    try:
        yield None
    finally:
        sys.stdout = save_stdout
        sys.stderr = save_stderr

def run_exectests(test_dir, log_path='exectests.log'):

    test_files = glob.glob(os.path.join(test_dir, '*.py'))
    test_files = [ex for ex in test_files if ex[0] is not '_']
    test_files.sort()
    passed = []
    failed = []
    with open(log_path, 'w') as f:
        with redirected_output(new_stdout=f, new_stderr=f):
            for fname in test_files:
                print(">> Executing '%s'" % fname)
                try:
                    code = compile(set_backend + open(fname, 'r').read(),
                                   fname, 'exec')
                    exec(code, {'__name__':'__main__'}, {})
                    passed.append(fname)
                except:
                    traceback.print_exc()
                    failed.append(fname)
                    pass

    print(">> Passed %i/%i tests: " %(len(passed), len(test_files)))
    print("Passed: " + ', '.join(passed))
    print("Failed: " + ', '.join(failed))
    print("See %s for details" % log_path)

    return passed, failed

if __name__ == '__main__':
    try:
        run_exectests(*sys.argv[1:])
    except:
        run_exectests(os.getcwd())





'''
Created on 20 mrt. 2013

@author: localadmin
'''
from __future__ import unicode_literals, absolute_import

from ema_workbench.connectors.netlogo import NetLogoModel

from ema_workbench import (RealParameter, ema_logging,
                           perform_experiments,TimeSeriesOutcome)
from ema_workbench.em_framework.evaluators import MultiprocessingEvaluator


if __name__ == '__main__':
    #turn on logging
    ema_logging.log_to_stderr(ema_logging.INFO)

    model = NetLogoModel('predprey', 
                          wd="./models/predatorPreyNetlogo", 
                          model_file="Wolf Sheep Predation.nlogo")
    model.run_length = 100
    model.replications = 10
    
    model.uncertainties = [RealParameter("grass-regrowth-time", 1, 99),
                           RealParameter("initial-number-sheep", 1, 200),
                           RealParameter("initial-number-wolves", 1, 200),
                           RealParameter("sheep-reproduce", 1, 20),
                           RealParameter("wolf-reproduce", 1, 20),
                     ]
    
    model.outcomes = [TimeSeriesOutcome('sheep'),
                      TimeSeriesOutcome('wolves'),
                      TimeSeriesOutcome('grass') ]
     
    #perform experiments
    n = 10
    
    with MultiprocessingEvaluator(model) as evaluator:
        results = perform_experiments(model, n, evaluator=evaluator)


'''
This module provides functionality for doing dimensional stacking of 
uncertain factors in order to reveal patterns in the values for a single 
outcome of interests. It is inspired by the work reported `here <https://www.onepetro.org/conference-paper/SPE-174774-MS>`_ 
with one deviation. 

Rather than using association rules to identify the 
uncertain factors to use, this code uses random forest based feature scoring
instead. It is also possible to use the code provided here in combination
with any other feature scoring or factor prioritization technique instead, or
by simply selecting uncertain factors in some other manner. 


'''
from __future__ import (division, unicode_literals, print_function,
                        absolute_import)

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from . import feature_scoring


# Created on Nov 13, 2015
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>


def discretize(data, nbins=3, with_labels=False):
    ''' Discretize the data, using the number of bins specified.
    
    Parameters
    ----------
    data : DataFrame
    nbins : int, optional
            the number of bins to use (default is 3)
    
    Returns
    -------
    digitized
        the digitized data frame
    
    note:: nbins is currently a constant for all float and integer columns. 
           Categorical data is not discretized. If the number of integers is
           lower than the number of bins, the integer variable is also not 
           discretized.
    

    '''
    discretized = data.copy()
    
    for i, entry in enumerate(data.dtypes):
        column = data.columns[i]
        column_data = data[column]
        n = nbins
        
        if issubclass(entry.type, np.integer):
            n_unique = column_data.unique().shape[0]
            if n_unique <= n:
                n = n_unique
        elif entry == np.dtype(object):
            n_unique = column_data.unique().shape[0]
            n = n_unique
            column_data= column_data.astype('category')
            column_data= column_data.cat.rename_categories([x for x in range(1,n+1)])
            
        if with_labels:
            labels = ['{}-{}'.format(column, i) for i in range(n)]
            indices = pd.cut(column_data, n,  
                             labels=labels)
        else:
            indices = pd.cut(column_data, n, retbins=False, labels=False)

        discretized[column] = indices
        
    return discretized


def dim_ratios(axis, figsize, side_colors_ratio=0.05):
    """Get the proportions of the figure taken up by each axes
    
    adapted from seaborn
    """
    figdim = figsize[axis]
    # Get resizing proportion of this figure for the dendrogram and
    # colorbar, so only the heatmap gets bigger but the dendrogram stays
    # the same size.
    dendrogram = min(2. / figdim, .2)

    # add the colorbar
    colorbar_width = .8 * dendrogram
    colorbar_height = .2 * dendrogram
    if axis == 0:
        ratios = [colorbar_width, colorbar_height]
    else:
        ratios = [colorbar_height, colorbar_width]

    # Add the ratio for the heatmap itself
    ratios += [.8]

    return ratios


def plot_line(ax, axis, i, lw, length):
    '''Helper function for plotting lines separating bins in the hierarchical
    index'''
    
    if axis==0:
        ax.plot([i,i], [length,1],lw=lw, color='grey')
    else:
        ax.plot([length,1],[i,i],lw=lw, color='grey')

        
def plot_category(ax, axis, i, label, pos):
    '''helper function for ploting label'''
    
    if axis==0:
        ax.text(i, pos, label, ha='center', va='center')
    else:
        ax.text(pos, i, label, ha='center', va='center')


def plot_index(ax, ax_plot, axis, index, plot_labels=True, plot_cats=True):
    '''helper function for visualizing the hierarchical index
    
    Parameters
    ----------
    
    ax : Axes instance
         the axes on which to plot the hierarchical index
    ax_plot : Axes instance
              the axes on which the table itself is displayed
    axis : int {0, 1}
           indicates whether we are plotting rows or columns
    plot_label : bool, optional
                 if true, also plot names of uncertain factors
    plot_cats : bool, options
                if true, plot category names for uncertain factors
    
    '''
    
    for entry in ['bottom', 'top', 'right', 'left']:
        ax.spines['bottom'].set_color('grey')
    
    if axis==0:
        names = index.names
        ax.spines['top'].set_color('white')
        ax.spines['top'].set_linewidth(1.0)
        ax.spines['bottom'].set_color('white')
        ax.spines['bottom'].set_linewidth(1.0)
        
        ax.invert_yaxis()
        ax.yaxis.tick_right()
        ax.set_xticks([])
        
        if plot_labels:
            tick_locs = np.linspace(1/(2*len(names)), 1-1/(2*len(names)),
                                    len(names))
            ax.set_yticks(tick_locs)
            ax.set_yticklabels(names)
        else:
            ax.set_yticks([])
    else:
        index = index[::-1]
        names = index.names
        ax.set_yticks([])
        
        ax.spines['left'].set_color('white')
        ax.spines['left'].set_linewidth(1.0)
        ax.spines['right'].set_color('white')
        ax.spines['right'].set_linewidth(1.0)
        
        
        if plot_labels:
            tick_locs = np.linspace(1/(2*len(names)), 1-1/(2*len(names)),
                                    len(names))
            ax.set_xticks(tick_locs)
            ax.set_xticklabels(names, rotation='vertical')
        else:
            ax.set_xticks([])

    try:
        nr_levels = len(index.levels)
        levels = index.levels
        indices = index.values
    except AttributeError:
        nr_levels = 1
        levels = [index.values.tolist()]
        indices = zip(index.values)

    last = indices[0]
    plot_line(ax, axis, 0, 1, 0) # first line
    plot_line(ax, axis, len(indices), 1, 0) #last line
    
    offsets = {}
    for i, level in enumerate(levels):
        offset = 1
        for entry in levels[0:i]:
            offset *= len(entry)

        offset *= len(level)
        offset = 1/(offset*2)

        offsets[i] = offset
        
    if plot_cats:
        for p in range(0, nr_levels):
            pos = 1/(2*nr_levels) + p /(nr_levels)
            plot_category(ax, axis, 0+offsets[p]*len(index), last[p], pos)

    for i, entry in enumerate(indices[1::]):
        i +=1
        comparison = map(lambda a, b: a==b, entry, last)

        for j, item in enumerate(comparison):
            if not item:
                ratio = j/nr_levels
                lw = 1 * (1-ratio)
                length = ratio
                break

        last = entry    
        plot_line(ax, axis, i, lw, length)

        if plot_cats:
            # add values
            for p in range(j, nr_levels):
                pos = 1/(2*nr_levels) + p /(nr_levels)
                plot_category(ax, axis, i+offsets[p]*len(index), entry[p], pos)
        if axis:
            ax_plot.axhline(i, c="w", lw=lw)
        else:
            ax_plot.axvline(i, c="w", lw=lw)        
        

def plot_pivot_table(table, plot_labels=True, plot_cats=True, 
                     figsize = (10, 10), cmap='viridis', **kwargs):
    ''' visualize a pivot table using colors
    
    Parameters
    ----------
    table : Pandas DataFrame
    plot_labels : bool, optional
                 if true, display uncertain factor names
    plot_cats : bool, optional
                 if true, display category labels for each uncertain factor
    fig_size : tuple of 2 ints, optional
               size of the figure to create
    cmap : matplotlib colormap name or object, optional
           default is viridis (requires matplotlib 1.5 or higher)
    kwargs : other keyword arguments
             All other keyword arguments are passed to ax.pcolormesh.
    
    Returns
    -------
    Figure
    
    '''
    
    with sns.axes_style('white'):
    
        fig = plt.figure(figsize=figsize)
    
        width_ratios = dim_ratios(figsize=figsize, axis=1)
        height_ratios = dim_ratios(figsize=figsize, axis=0)
    
        gs = mpl.gridspec.GridSpec(3, 3, wspace=0.01, hspace=0.01, 
                                  width_ratios=width_ratios,
                                  height_ratios=height_ratios)
    
        ax_plot = fig.add_subplot(gs[2,2])
        ax_rows = fig.add_subplot(gs[2, 0:2], axisbg="white")
        ax_cols = fig.add_subplot(gs[0:2, 2], axisbg="white")
        cax = fig.add_subplot(gs[0, 0])
    
        # actual plotting
        plot_data = table.values
        sns.heatmap(plot_data, ax=ax_plot, cbar_ax=cax, cmap=cmap, **kwargs)
    
        # set the tick labels
        ax_plot.set_xticks([])
        ax_plot.set_yticks([])
    
        # plot row labeling
        ax_rows.set_ylim(ax_plot.get_ylim())
        ax_rows.set_xlim(0,1)
        plot_index(ax_rows, ax_plot, axis=1, index=table.index,
                   plot_labels=plot_labels, plot_cats=plot_cats)
    
        # plot column labeling
        ax_cols.set_xlim(ax_plot.get_xlim())
        ax_cols.set_ylim(0,1)
        plot_index(ax_cols, ax_plot, axis=0, index=table.columns,
                   plot_labels=plot_labels, plot_cats=plot_cats)
    
    
    return fig


def make_pivot_table(data, rows=None, columns=None, values=None):
    ''' Make a pivot table
    
    work around for fact that dropna=False kwarg on pivot tables
    results in column and row names being dropped
    
    Parameters
    ----------
    data : DataFrame
    rows : iterable of str
    columns : iterable of str
    values : str
    
    Returns
    -------
    DataFrame
    
    '''
    
    stacked = pd.pivot_table(data, values=values, index=rows, 
              columns=columns, dropna=False)    
    stacked.index.names = rows
    stacked.columns.names = columns
    
    return stacked


def create_pivot_plot(x, y, nr_levels=3, labels=True, categories=True,
                     nbins=3):
    ''' convenience function for easily creating a pivot plot
    
    Parameters
    ----------
    x : structured array
    y : 1d ndarray
    nr_levels : int, optional
                the number of levels in the pivot table. The number of 
                uncertain factors included in the pivot table is two times the 
                number of levels.
    labels : bool, optional
             display names of uncertain factors
    categories : bool, optional
                 display category names for each uncertain factor
    nbins : int, optional
            number of bins to use when discretizing continuous uncertain 
            factors
    
    Returns
    -------
    Figure
    
    
    This function performs feature scoring using random forests, selects a 
    number of high scoring factors based on the specified number of levels, 
    creates a pivot table, and visualizes the table. This is a convenience 
    function. For more control over the process, use the code in this function 
    as a template. 
        
    '''
    scores = feature_scoring.get_ex_feature_scores(x,y)[0]
    
    n = nr_levels*2
    
    scores = scores[0].values.tolist()
    rows = [entry for entry in scores[0:n:2]]
    columns = [entry for entry in scores[1:n:2]]

    data = pd.DataFrame.from_records(x)
    discretized_x = discretize(data, nbins=3)
    
    ooi_label = 'y'
    ooi = pd.DataFrame(y[:, np.newaxis], columns=[ooi_label])
    
    x_y_concat = pd.concat([discretized_x, ooi], axis=1)
    pvt = make_pivot_table(x_y_concat, rows=rows, columns=columns, 
                           values=ooi_label)

    fig = plot_pivot_table(pvt, plot_labels=labels, plot_cats=categories)
    
    return fig


'''
Created on Nov 24, 2011

@author: gonengyucel
'''
from __future__ import print_function

import numpy as np
from ema_workbench.util import info


def trdist(d1,d2):
    
    a = np.sum(d1*d2)
    b = np.sqrt(np.sum(d1*d1))
    c = np.sqrt(np.sum(d2*d2))
    d = a/(b*c)
    print(d)
    return d

def distance_triangle(data):
    '''
    The triangle distance is calculated as follows;
        Let ds1(.) and ds2(.) be two data series of length N. Then;
        A equals to the summation of ds1(i).ds2(i) from i=1 to N
        B equals to the square-root of the (summation ds1(i)^2 from i=1 to N)
        C equals to the square-root of the (summation ds1(i)^2 from i=1 to N)
        
        distance_triangle = A/(B.C)
     
     The triangle distance works only with data series of the same length
     
     In the literature, it is claimed that the triangle distance can deal with noise and amplitude scaling very well, and may yield poor
     results in cases of offset translation and linear drift.   
    '''
    
    
    
    runLogs = []
    #Generates the feature vectors for all the time series that are contained in numpy array data
    info("calculating distances")
    dRow = np.zeros(shape=(np.sum(np.arange(data.shape[0])), ))
    index = -1
    for i in range(data.shape[0]):
            
        # For each run, a log is created
        # Log includes a description dictionary that has key information 
        # for post-clustering analysis, and the data series itself. These 
        # logs are stored in a global array named runLogs
        behaviorDesc = {}
        behaviorDesc['Index'] = str(i)
        
        behavior = data[i]
        localLog = (behaviorDesc, behavior)
        runLogs.append(localLog)
    
        for j in range(i+1, data.shape[0]):
            index += 1
            distance = trdist(data[i],data[j]) 
            dRow[index] = distance
    return dRow, runLogs

if __name__ == '__main__':
    tester = np.array([(1,2,3,4,5,6,7,8,12,15),(11,22,32,32,26,24,23,16,12,10)])



'''

Feature scoring functionality 


'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

from operator import itemgetter

import numpy as np
import numpy.lib.recfunctions as recfunctions
import pandas as pd
import six
from sklearn.ensemble import (ExtraTreesClassifier, ExtraTreesRegressor,
                              RandomForestClassifier, RandomForestRegressor)
from sklearn.feature_selection.univariate_selection import (f_regression,
                                                            f_classif, chi2)
from sklearn.linear_model.least_angle import LassoLarsCV
from sklearn.linear_model.randomized_l1 import (RandomizedLogisticRegression,
                                                RandomizedLasso)

from .scenario_discovery_util import CLASSIFICATION, REGRESSION

# Created on Jul 9, 2014
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
#
# TODO:: look at http://scikit-learn.org/stable/auto_examples/linear_model/plot_sparse_recovery.html#example-linear-model-plot-sparse-recovery-py


__al__ = ['F_REGRESSION', 'F_CLASSIFICATION', 'CHI2', 
          'get_univariate_feature_scores', 'get_rf_feature_scores',
          'get_lasso_feature_scores', 'get_feature_scores_all']

F_REGRESSION = f_regression


F_CLASSIFICATION = f_classif

CHI2 = chi2


def _prepare_experiments(experiments):
    '''
    transform the experiments structured array into a numpy array.

    Parameters
    ----------
    experiments : structured array
    
    Returns
    -------
    ndarray
    
    '''
    experiments = recfunctions.drop_fields(experiments, "scenario_id", 
                                           asrecarray=True)
    uncs = recfunctions.get_names(experiments.dtype)

    temp_experiments = np.zeros((experiments.shape[0], len(uncs)))
    
    for i, u in enumerate(uncs):
        try: 
            temp_experiments[:,i] = experiments[u].astype(np.float)
        except ValueError:
            
            data = experiments[u]
            entries = sorted(list(set(data)))
            
            for j, entry in enumerate(entries):
                temp_experiments[data==entry,i] = j
    
    return temp_experiments, uncs


def _prepare_outcomes(outcomes, classify):
    '''
    transform the outcomes dict into a vector with either the class allocation
    or the value.
    
    Parameters
    ----------
    outcomes : dict 
               the outcomes dict
    classify : callable or str
               a classify function or variable analogous to PRIM
               
    Returns
    -------
    1d ndarray 
        the return from classify
    bool
        data is categorical (True) or continuous (False)
    
    Raises
    --------
    TypeError 
        if classify is neither a StringType nor a callable
    KeyError 
        if classify is a string which is not a key in the outcomes dict.
    
    '''
    if isinstance(classify, six.string_types):
        try:
            y = outcomes[classify]
        except KeyError as e:
            raise e
        categorical = False
    elif callable(classify):
        y = classify(outcomes)
        categorical=True
    else:
        raise TypeError("unknown type for classify")
    
    return y, categorical


def get_univariate_feature_scores(x,y, score_func=F_CLASSIFICATION):
    '''
    
    calculate feature scores using univariate statistical tests. In case of
    categorical data, chi square or the Anova F value is used. In case of 
    continuous data the Anova F value is used. 
    
    Parameters
    ----------
    x : structured array
    y : 1D nd.array
    score_func : {F_CLASSIFICATION, F_REGRESSION, CHI2}
                the score function to use, one of f_regression (regression), or  
                f_classification or chi2 (classification). 
    Returns
    -------
    pandas DataFrame
        sorted in descending order of tuples with uncertainty and feature 
        scores (i.e. p values in this case).
    
    
    '''
    x, uncs = _prepare_experiments(x)
    
    pvalues = score_func(x, y)[1]
    pvalues = np.asarray(pvalues)

    pvalues = zip(uncs, pvalues)
    pvalues = list(pvalues)
    pvalues.sort(key=itemgetter(1))
    
    pvalues = pd.DataFrame(pvalues)
    
    return pvalues


def get_rf_feature_scores(x, y, mode=CLASSIFICATION, nr_trees=250, 
                          max_features='auto', max_depth=None, 
                          min_samples_split=2, min_samples_leaf=1, 
                          bootstrap=True, oob_score=True, random_state=None): 
    '''
    Get feature scores using a random forest

    Parameters
    ----------
    x : structured array
    y : 1D nd.array
    mode : {CLASSIFICATION, REGRESSION}
    nr_trees : int, optional
               nr. of trees in forest (default=250)
    max_features : int, optional
                   see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    max_depth : int, optional 
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    min_samples : int, optional
                  see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    min_samples_leaf : int, optional
                       see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    bootstrap : bool, optional
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    oob_score : bool, optional
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    random_state : int, optional
                   see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    
    Returns
    -------
    pandas DataFrame
        sorted in descending order of tuples with uncertainty and feature 
        scores 
    object
        either RandomForestClassifier or RandomForestRegressor
    
    '''
    x, uncs = _prepare_experiments(x)
    
    if mode==CLASSIFICATION:
        rfc = RandomForestClassifier
        criterion='gini'
    elif mode==REGRESSION:
        rfc = RandomForestRegressor
        criterion = 'mse'
    else:
        raise ValueError('{} not valid for mode'.format(mode))
    
    forest = rfc(n_estimators=nr_trees, 
                criterion=criterion, 
                max_features=max_features, 
                max_depth=max_depth,
                min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                bootstrap=bootstrap,
                oob_score=oob_score,
                random_state=random_state)
    forest.fit(x,y)

    importances = forest.feature_importances_

    importances = zip(uncs, importances)
    importances = list(importances)
    importances.sort(key=itemgetter(1), reverse=True)

    importances = pd.DataFrame(importances)

    return importances, forest


def get_lasso_feature_scores(x, y, mode=CLASSIFICATION, scaling=0.5, 
                             sample_fraction=0.75, n_resampling=200,
                             random_state=None):
    '''
    Calculate features scores using a randomized lasso (regression) or 
    randomized logistic regression (classification). This is also known as 
    stability selection.
    
    see http://scikit-learn.org/stable/modules/feature_selection.html for 
    details. 
    
    Parameters
    ----------   
    x : structured array
    y : 1D nd.array
    mode : {CLASSIFICATION, REGRESSION}
    scaling : float, optional
              scaling parameter, should be between 0 and 1
    sample_fraction : float, optional
                      the fraction of samples to used in each randomized 
                      dataset
    n_resmpling : int, optional
                  the number of times the model is trained on a random subset 
                  of the data
    random_state : int, optional
                   if it is an int, it specifies the seed to use, defaults to 
                   None.
                         
    Returns
    -------
    pandas DataFrame
        sorted in descending order of tuples with uncertainty and feature 
        scores         
         
    '''
    x, uncs = _prepare_experiments(x)
    
    if mode==CLASSIFICATION:

        lfs = RandomizedLogisticRegression(scaling=scaling, 
                                           sample_fraction=sample_fraction,
                                           n_resampling=n_resampling, 
                                           random_state=random_state)
        lfs.fit(x,y)
    elif  mode==REGRESSION:
        # we use LassoLarsCV to determine alpha see
        # http://scikit-learn.org/stable/auto_examples/linear_model/plot_sparse_recovery.html
        lars_cv = LassoLarsCV(cv=6).fit(x, y,)
        alphas = np.linspace(lars_cv.alphas_[0], .1 * lars_cv.alphas_[0], 6)
        
        # fit the randomized lasso        
        lfs = RandomizedLasso(alpha=alphas,scaling=scaling, 
                              sample_fraction=sample_fraction,
                              n_resampling=n_resampling,
                              random_state=random_state)
        lfs.fit(x, y)
    else:
        raise ValueError('{} invalid value for mode'.format(mode))

    importances = lfs.scores_
    importances = zip(uncs, importances)
    importances = list(importances)
    importances.sort(key=itemgetter(1), reverse=True)
    importances = pd.DataFrame(importances)


    return importances

def get_ex_feature_scores(x, y, mode=CLASSIFICATION, nr_trees=250, 
                          max_features='auto', max_depth=None, 
                          min_samples_split=2, min_samples_leaf=1, 
                          min_weight_fraction_leaf=0, max_leaf_nodes=None,
                          bootstrap=True, oob_score=True, random_state=None): 
    '''
    Get feature scores using extra trees

    Parameters
    ----------
    x : structured array
    y : 1D nd.array
    mode : {CLASSIFICATION, REGRESSION}
    nr_trees : int, optional
               nr. of trees in forest (default=250)
    max_features : int, optional
                   see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    max_depth : int, optional 
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    min_samples_split : int, optional
                  see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    min_samples_leaf : int, optional
                       see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    min_weight_fraction_leaf : float, optional
                               see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    max_leaf_nodes: int or None, optional
                    see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    bootstrap : bool, optional
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    oob_score : bool, optional
                see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    random_state : int, optional
                   see http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    
    Returns
    -------
    pandas DataFrame
        sorted in descending order of tuples with uncertainty and feature 
        scores 
    object
        either ExtraTreesClassifier or ExtraTreesRegressor
    
    '''
    x, uncs = _prepare_experiments(x)
    
    if mode==CLASSIFICATION:
        etc = ExtraTreesClassifier
        criterion='gini'
    elif mode==REGRESSION:
        etc = ExtraTreesRegressor
        criterion = 'mse'
    else:
        raise ValueError('{} not valid for mode'.format(mode))
    
    extra_trees = etc(n_estimators=nr_trees, 
                      criterion=criterion, 
                      max_features=max_features, 
                      max_depth=max_depth,
                      min_samples_split=min_samples_split,
                      min_samples_leaf=min_samples_leaf,
                      min_weight_fraction_leaf=min_weight_fraction_leaf,
                      max_leaf_nodes=max_leaf_nodes,
                      bootstrap=bootstrap,
                      oob_score=oob_score,
                      random_state=random_state)
    extra_trees.fit(x,y)

    importances = extra_trees.feature_importances_

    importances = zip(uncs, importances)
    importances = list(importances)
    importances.sort(key=itemgetter(1), reverse=True)

    importances = pd.DataFrame(importances)

    return importances, extra_trees

algorithms = {'extra trees' : get_ex_feature_scores,
              'lasso' : get_lasso_feature_scores,
              'random forest' : get_rf_feature_scores,
              'univariate' : get_univariate_feature_scores}

def get_feature_scores_all(x, y, alg='extra trees', mode=REGRESSION,
                   **kwargs):
    '''perform feature scoring for all outcomes using the specified feature 
    scoring algorithm
    
    Parameters
    ----------
    x : numpy structured array
    y : dict of 1d numpy arrays
        the outcomes, with a string as key, and a 1D array for each outcome
    alg : {'extra trees', 'lasso', 'random forest', 'univariate'}, optional
    mode : {REGRESSION, CLASSIFICATION}, optional
    kwargs : dict, optional
             any remaining keyword arguments will be passed to the specific
             feature scoring algorithm
    
    Returns
    -------
    DataFrame instance
    
    
    '''
    complete = None
    for key, value in y.items():
        fs, _ = algorithms[alg](x,value,mode=mode, **kwargs)

        fs = fs.rename(columns={1:key})
        fs = fs.set_index(0)

        if complete is None:
            complete = fs.T
        else:
            complete = complete.append(fs.T)

    return complete.T




'''
helper module for running experiments and keeping track of which model
has been initialized with which policy. 
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import sys
import traceback

from ..util import ema_logging, EMAError, CaseError

# Created on Aug 11, 2015
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ["ExperimentRunner"]

class ExperimentRunner(object):
    '''Helper class for running the experiments
    
    This class contains the logic for initializing models properly,
    running the experiment, getting the results, and cleaning up afterwards.
    
    Parameters
    ----------
    msis : dict
    model_kwargs : dict
    
    Attributes
    ----------
    msi_initializiation : dict
                          keeps track of which model is initialized with
                          which policy. 
    msis : dict
           models indexed by name
    model_kwargs : dict
                   keyword arguments for model_init
    
    '''
    
    def __init__ (self, msis):
        self.msis = msis
        self.log_message = ('running scenario {scenario_id} for policy '
                            '{policy_name} on model {model_name}')
    
    def cleanup(self):
        for msi in self.msis:
            msi.cleanup()
        self.msis = None
    
    def run_experiment(self, experiment):
        '''The logic for running a single experiment. This code makes
        sure that model(s) are initialized correctly.
        
        Parameters
        ----------
        experiment : dict
        
        Returns
        -------
        experiment_id: int
        case : dict
        policy : str
        model_name : str
        result : dict
        
        Raises
        ------
        EMAError
            if the model instance raises an EMA error, these are reraised.
        Exception
            Catch all for all other exceptions being raised by the model. 
            These are reraised.
        
        '''
        
        policy_name = experiment.policy.name
        model_name = experiment.model_name
        model = self.msis[model_name]
        policy = experiment.policy.copy()
        scenario_id = experiment.scenario.name
        
        ema_logging.debug(self.log_message.format(scenario_id=scenario_id,
                                                  policy_name=policy_name,
                                                  model_name = model_name))
        scenario = experiment.scenario.copy()
        try:
            model.run_model(scenario, policy)
        except CaseError as e:
            ema_logging.warning(str(e))
        except Exception as e:
            try:
                self.cleanup()
            except Exception:
                raise e
            
            exception = traceback.print_exc()
            if exception:
                sys.stderr.write(exception)
                sys.stderr.write("\n")
            
            raise EMAError(("exception in run_model"
                   "\nCaused by: {}: {}".format(type(e).__name__, str(e))))
            
        output = model.output
        model.reset_model()
        
        return output      


'''

a reworking of the cluster. The distance metrics have now their own .py file. 
The metrics available are currently stored in the distance_functions 
dictionary.

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import matplotlib.pyplot as plt
import numpy as np
from scipy.cluster.hierarchy import fcluster, linkage, dendrogram, inconsistent
from scipy.spatial.distance import squareform

from .cluster_util import (distance_mse, distance_sse, distance_triangle,
                           distance_gonenc)
from ..util import info, EMAError

# Created on Sep 8, 2011
# 
# .. codeauthor:: 
#      gyucel <g.yucel (at) tudelft (dot) nl>,
#      jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
#                 

distance_functions = {'gonenc': distance_gonenc,
                      'triangle':distance_triangle,
                      'sse': distance_sse,
                      'mse': distance_mse}

# Global variables
runLogs = []
varName = ""
clusterCount = 0

def cluster(data, 
            outcome,
            distance='gonenc',
            interClusterDistance='complete',
            cMethod='inconsistent',
            cValue=2.5,
            plotDendrogram=True,
            plotClusters=True,
            groupPlot=False,
            **kwargs):
    '''
    
    Method that clusters time-series data from the specified cpickle file 
    according to a selected distance measure.
    
    Parameters
    ----------
    
    data : tuple
           return from meth:`perform_experiments`.
    outcome : str
              Name of outcome/variable whose behavior is being analyzed
    distance : {'gonenc','triangle', 'sse', 'mse'}
               The distance metric to be used.
    interClusterDistance : str
                           How to calculate inter cluster distance.
                           see `linkage <http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html#scipy.cluster.hierarchy.linkage>`_ 
                           for details.
    cMethod : str
              Cutoff method, see `fcluster <http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.fcluster.html#scipy.cluster.hierarchy.fcluster>`_ 
              for details.
    cValue : float
             Cutoff value, see `fcluster <http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.fcluster.html#scipy.cluster.hierarchy.fcluster>`_ 
             for details.
    plotDendogram : bool
    plotCluster : bool
    groupPlot: bool
    
    Returns
    -------
    list
        distances
    list
        Clusters
    list
        distance metrics
    
    The remainder of the arguments are passed on to the specified distance 
    function.
    
    Gonenc Distance:
    
    * 'distance': String that specifies the distance to be used. 
                  Options: bmd (default), mse, sse
    * 'filter?': Boolean that specifies whether the data series will be 
                 filtered (for bmd distance)
    * 'slope filter': A float number that specifies the filtering threshold 
                     for the slope (for every data point if change__in_the_
                     outcome/average_value_of_the_outcome < threshold, 
                     consider slope = 0) (for bmd distance)
    * 'curvature filter': A float number that specifies the filtering 
                          threshold for the curvature (for every data point if 
                          change__in_the_slope/average_value_of_the_slope < 
                          threshold, consider curvature = 0) (for bmd distance)
    * 'no of sisters': 50 (for bmd distance)

    '''
    
    global varName 
    varName = outcome
    data = data[1][outcome]
    
    # Construct a list with distances. This list is the upper triange
    # of the distance matrix
    dRow, runLogs = construct_distances(data, distance, **kwargs)
    info('finished distances')
    
    # Allocate individual runs into clusters using hierarchical agglomerative 
    # clustering. clusterSetup is a dictionary that customizes the clustering 
    # algorithm to be used.
    z, clusters, runLogs = flatcluster(dRow, 
                                    runLogs, 
                                    plotDendrogram=plotDendrogram,
                                    interClusterDistance=interClusterDistance,
                                    cMethod=cMethod,
                                    cValue=cValue)
    
    info("tranforming to list of clusters")
    clusters = make_data_structure(clusters, dRow, runLogs)

    if plotClusters:
        plot_clusters(groupPlot, runLogs)
    
    return dRow, clusters, z

def make_data_structure(clusters, distRow, runLogs):
    nr_clusters = np.max(clusters)
    cluster_list = []
    for i in range(1, nr_clusters+1):
        info("starting with cluster %s" %i)
        #determine the indices for cluster i
        indices = np.where(clusters==i)[0]
        
        drow_indices = np.zeros((indices.shape[0]**2-indices.shape[0])/2, dtype=int)
        s = 0
        #get the indices for the distance for the runs in the cluster
        for q in range(indices.shape[0]):
            for r in range(q+1, indices.shape[0]):
                b = indices[q]
                a = indices[r]
                
                drow_indices[s] = get_drow_index(indices[r],
                                                 indices[q], 
                                                 clusters.shape[0])
                s+=1
        
        #get the distance for the runs in the cluster
        dist_clust = distRow[drow_indices]
        
        #make a distance matrix
        dist_matrix = squareform(dist_clust)

        #sum across the rows
        row_sum = dist_matrix.sum(axis=0)
        
        #get the index of the result with the lowest sum of distances
        min_cIndex = row_sum.argmin()
    
        # convert this cluster specific index back to the overall cluster list 
        # of indices
        originalIndices = np.where(clusters==i)
        originalIndex = originalIndices[0][min_cIndex]

        print(originalIndex)

        a = list(np.where(clusters==i)[0])
        a = [int(entry) for entry in a]
        
        cluster = Cluster(i, 
                          np.where(clusters==i)[0], 
                          originalIndex,
                          [runLogs[entry] for entry in a],
                          dist_clust)
        cluster_list.append(cluster)
    return cluster_list

def get_drow_index(i,j, size):
    '''
    Get the index in the distance row for the distance between i and j.
    
    Parameters
    ----------
    i : result i
    j : result j
    size : the number of results
    
    Returns
    -------
    int
    
    
    ...note:: i > j
    
    '''
    assert i > j

    index = 0
    for q in range(size-j, size):
        index += q
    index = index+(i-(1*j))-1

    return index

def construct_distances(data, distance='gonenc', **kwargs):
    """ 
        
    Constructs a n-by-n matrix of distances for n data-series in data 
    according to the specified distance.
    
    Distance argument specifies the distance measure to be used. Options, 
    which are defined in clusteringDistances.py, are as follows.
    
    
    * gonenc: a distance based on qualitative dynamic pattern features 
    * willem: a disance mainly based on the presence of crisis-periods and 
              the overall trend of the data series
    * sse: regular sum of squared errors
    * mse: regular mean squared error
    
    
    SSE and MSE are in clusterinDistances.py and don't work right now.
    
    others will be added over time
    
    """
    
    # Sets up the distance function according to user specification
    try:
        return distance_functions[distance](data, **kwargs)
    except KeyError:
        raise EMAError("trying to use an unknown distance: %s" %(distance))

def flatcluster(dRow, runLogs, 
                interClusterDistance='complete',
                plotDendrogram=True,
                cMethod='inconsistent',
                cValue=2.5):
#    if 'inter-cluster distance' in clusterSetup.keys():
#        method = clusterSetup['inter-cluster distance']
#    else:
#        method = 'complete'

    z = linkage(dRow, interClusterDistance)
    inc = inconsistent(z)
    #print inc
    
    if plotDendrogram:
        plotdendrogram(z)
    
    clusters = fcluster(z, cValue, cMethod)
    
    noClusters = max(clusters)
    print('Total number of clusters:', noClusters)
    for i in range(noClusters):
        counter = 0
        for j in range(len(clusters)):
            if clusters[j]==(i+1):
                counter+=1
        print("Cluster",str(i+1),":",str(counter))
    
    global clusterCount
    clusterCount = noClusters
    print(len(clusters))
    print(len(runLogs))
    for i, log in enumerate(runLogs):
        log[0]['Cluster'] = str(clusters[i])
    

    return z, clusters, runLogs
           
def plotdendrogram(z):
    
    dendrogram(z,
               truncate_mode='lastp',
               show_leaf_counts=True,
               show_contracted=True
               )
    plt.show()

def plot_clusters(groupPlot, runLogs):
    global varName
    global clusterCount
    noRuns = len(runLogs)
    
    clustersToPlot = []
    #print 'Cluster Sayisi ', clusterCount
    # For each cluster, create a data structure
    for clust in range(clusterCount):    
        runSubset = []
        for runIndex in range(noRuns):
            if runLogs[runIndex][0]['Cluster']==str(clust+1):
                runSubset.append(runLogs[runIndex])
        
   
class Cluster(object):
    '''
    Contains information about a data-series cluster, as well as some methods to help analyzing a cluster.
    Basic attributes of a cluster (e.g. c) object are as follows;
     
     * c.no : Cluster number/index
     * c.indices : Original indices of the dataseries that are in cluster c
     * c.sample : Original index of the dataseries that is the representative of cluster c (i.e. median element of the cluster)
     * c.size : Number of elements (i.e. dataseries) in the cluster c
    '''

    def __init__(self, 
                 cluster_no, 
                 all_ds_indices, 
                 sample_ds_index,
                 runLogs,
                 dist_clust):
        '''
        Constructor
        '''
        self.no = cluster_no
        self.indices = all_ds_indices
        self.sample = int(sample_ds_index)
        self.size = self.indices.size
        self.runLogs = runLogs
        self.distances = dist_clust
        
        
    def error(self):
        return self.sample

if __name__ == '__main__':
    from ..util import load_results
    results = load_results('../sandbox/cluster/datasets/PatternSet_Basics.cPickle')
    
    
    distance, liste, obekler, kosulog, zet = cluster(results, 'outcome',cMethod='distance', cValue=1, groupPlot=True, plotDendrogram=False)
    
    #clusters = np.array([3, 1, 2, 2, 1,2])
    #dRow = np.array([1,2,3,6,2,1,0,3,9,6,2,1,3,2,1])
    #samples = pick_cSamples(clusters, dRow)
    #print "ornekler", samples
 
    
        
    





from __future__ import (absolute_import)
from . import analysis
from . import em_framework
from .em_framework import (Model, RealParameter, CategoricalParameter, 
                           IntegerParameter, perform_experiments, optimize,
                           ScalarOutcome, TimeSeriesOutcome, Constant, 
                           Scenario, Policy, MultiprocessingEvaluator, 
                           IpyparallelEvaluator)

from . import util
from .util import save_results, load_results, ema_logging, EMAError

__version__ = '0.9.4'


from __future__ import absolute_import

from .gonenc_distance import distance_gonenc
from .mse_distance import distance_mse
from .sse_distance import distance_sse
from .triangle_distance import distance_triangle



'''

This module offers basic functionality for converting a matplotlib figure
to black and white. The provided functionality is largely determined by
what is needed for the workbench. 

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import itertools
import math
import six

import matplotlib as mpl
import numpy as np
from matplotlib.collections import PolyCollection, PathCollection
from matplotlib.colors import ColorConverter

from ema_workbench.util import ema_logging
from ..util import EMAError


# Created on 18 sep. 2012
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['set_fig_to_bw']

bw_mapping = [{'marker': None, 'dash': (None,None), 'fill':'0.1', 'hatch':'/'},
              {'marker': None, 'dash': [5,5], 'fill':'0.25', 'hatch':'\\'},
              {'marker': None, 'dash': [5,3,1,3], 'fill':'0.4', 'hatch':'|'},
              {'marker': None, 'dash': [1,3], 'fill':'0.55', 'hatch':'-'},
              {'marker': None, 'dash': [5,2,5,2,5,10], 'fill':'0.7', 'hatch':'o'},
              {'marker': None, 'dash': [5,3,1,2,1,10], 'fill':'0.85', 'hatch':'O'},
              {'marker': 'o', 'dash': (None,None), 'fill':'0.1', 'hatch':'.'}]

MARKERSIZE = 3
HATCHING = 'hatching'
GREYSCALE = 'grey_scale'

def _identify_colors(fig):
    '''Identify the various colors that are used in the figure and
    return as a set
    
    '''
    
    color_converter = ColorConverter()
    all_colors = set()
    
    for ax in fig.axes:
        for line in ax.get_lines():
            orig_color = line.get_color()
            all_colors.add(orig_color)
        
        for patch in ax.patches:
            rgb_orig = color_converter.to_rgb(patch._facecolor)
            all_colors.add(rgb_orig)
                
        for collection in ax.collections:
            for color in collection.get_facecolor():
                rgb_orig = color_converter.to_rgb(color)
                all_colors.add(rgb_orig)
    
    return all_colors

def set_ax_lines_bw(ax, colormap, line_style='continuous'):
    """
    Take each Line2D in the axes, ax, and convert the line style to be 
    suitable for black and white viewing.
    
    Derived from and expanded for use in the EMA workbench from:
    http://stackoverflow.com/questions/7358118/matplotlib-black-white-colormap-with-dashes-dots-etc
    
    Parameters
    ----------
    ax : axes
         The ax of which the lines needs to be transformed to B&W. Lines are 
         transformed to different line styles.
    colormap : dict 
    line_style: str
                linestyle to use for converting, can be continuous, black
                or None
                    
    """

    for line in ax.get_lines():
        orig_color = line.get_color()
        try:
            mapping = colormap[orig_color]
        except:
            ema_logging.warning('no mapping specified for color: {}'.format(orig_color))
        else:
            if line_style == 'continuous':
                line.set_color('black')
                alpha = 1/(math.log(len(ax.get_lines()))+1)
                line.set_alpha(alpha)
            elif line_style == 'black':
                line.set_color('black')
                line.set_alpha(0.5)
            else:
                line.set_color('black')
                line.set_dashes(mapping['dash'])
                line.set_marker(mapping['marker'])
                line.set_alpha(1)
                line.set_markersize(MARKERSIZE)


def set_ax_patches_bw(ax, colormap):
    """
    Take each patch in the axes, ax, and convert the face color to be 
    suitable for black and white viewing.
    
    Parameters
    ----------
    ax : axes
         The ax of which the patches needs to be transformed to B&W.
    colormap : dict
               mapping of color to B&W rendering
    
    """    
    
    color_converter = ColorConverter()
    
    for patch in ax.patches:
        rgb_orig = color_converter.to_rgb(patch._facecolor)
        new_color = color_converter.to_rgba(colormap[rgb_orig]['fill'])
        
        patch._facecolor = new_color


def set_ax_collections_to_bw(ax, style, colormap):
    """
    Take each polycollection in the axes, ax, and convert the face color to be 
    suitable for black and white viewing.

    Parameters
    ----------
    ax : axes
        The ax of which the polycollection needs to be transformed to B&W.
    style : {HATCHING, GREYSCALE}
    colormap : dict
               mapping of color to B&W rendering
    
    """
    for collection in ax.collections:
        collection_type = type(collection).__name__
        try:
            converter_func = _collection_converter[collection_type]
        except KeyError:
            raise EMAError("converter for {} not implemented".format(collection_type))
        else:
            converter_func(collection, ax, style, colormap)
        
 
def _set_ax_polycollection_to_bw(collection, ax, style, colormap):
    '''helper function for converting a polycollection to black and white
    
    Parameters
    ----------
    collection : polycollection
    ax : axes
    style : {GREYSCALE, HATCHING}
    colormap : dict
               mapping of color to B&W rendering

    
    '''

    if style==GREYSCALE:
        color_converter = ColorConverter()
        for polycollection in ax.collections:
            orig_color = polycollection._original_facecolor

            try:
                mapping = colormap[orig_color]
            except:
                ema_logging.warning('no mapping specified for color: {}'.format(orig_color))
            else:
                new_color = color_converter.to_rgba(mapping['fill'])
                new_color = np.asarray([new_color])
                polycollection.update({'facecolors' : new_color}) 
                polycollection.update({'edgecolors' : new_color})
    elif style==HATCHING:
        orig_color = collection._original_facecolor
        
        try:
            mapping = colormap[orig_color]
        except:
            ema_logging.warning('no mapping specified for color: {}'.format(orig_color))
        else:
            collection.update({'facecolors' : 'none'}) 
            collection.update({'edgecolors' : 'white'}) 
            collection.update({'alpha':1})
        
            for path in collection.get_paths():
                p1 = mpl.patches.PathPatch(path, fc="none", 
                                           hatch=colormap[orig_color]['hatch'])
                ax.add_patch(p1)
                p1.set_zorder(collection.get_zorder()-0.1)


def _set_ax_pathcollection_to_bw(collection, ax, style, colormap):
    '''helper function for converting a pathcollection to black and white
    
    Parameters
    ----------
    collection : pathcollection
    ax : axes
    style : {GREYSCALE, HATCHING}
    colormap : dict
               mapping of color to B&W rendering
    
    '''
    color_converter = ColorConverter()
    colors = {}
    for key, value in color_converter.colors.items():
        colors[value] = key    

    rgb_orig = collection._original_facecolor

    if isinstance(rgb_orig, six.string_types):
        rgb_orig = [rgb_orig]
    rgb_orig = [color_converter.to_rgb(row) for row in rgb_orig]
    
    new_color = [color_converter.to_rgba(colormap[entry]['fill']) for entry 
                 in rgb_orig]
    new_color = np.asarray(new_color)
    
    collection.update({'facecolors' : new_color}) 
    collection.update({'edgecolors' : new_color}) 


_collection_converter = {PathCollection.__name__: _set_ax_pathcollection_to_bw,  # @UndefinedVariable
                         PolyCollection.__name__: _set_ax_polycollection_to_bw}  # @UndefinedVariable


def set_legend_to_bw(leg, style, colormap, line_style='continuous'):
    """
    Takes the figure legend and converts it to black and white. Note that
    it currently only converts lines to black and white, other artist 
    intances are currently not being supported, and might cause errors or
    other unexpected behavior.
    
    Parameters
    ----------
    leg : legend
    style : {GREYSCALE, HATCHING}
    colormap : dict
               mapping of color to B&W rendering
    line_style: str
                linestyle to use for converting, can be continuous, black
                or None
                
    """
    color_converter = ColorConverter()

    if leg:
        if isinstance(leg, list):
            leg = leg[0]
    
        for element in leg.legendHandles:
            if isinstance(element, mpl.collections.PathCollection):
                rgb_orig = color_converter.to_rgb(element._facecolors[0])
                new_color = color_converter.to_rgba(colormap[rgb_orig]['fill'])
                element._facecolors = np.array((new_color,))
            elif isinstance(element, mpl.patches.Rectangle):
                rgb_orig = color_converter.to_rgb(element._facecolor)
                
                if style==HATCHING:
                    element.update({'alpha':1})
                    element.update({'facecolor':'none'})
                    element.update({'edgecolor':'black'})
                    element.update({'hatch':colormap[rgb_orig]['hatch']})
                elif style==GREYSCALE:
                    ema_logging.info(colormap.keys())
                    element.update({'facecolor':colormap[rgb_orig]['fill']})
                    element.update({'edgecolor':colormap[rgb_orig]['fill']})
            else:
                line = element
                orig_color = line.get_color()
                
                line.set_color('black')
                if not line_style=='continuous':
                    line.set_dashes(colormap[orig_color]['dash'])
                    line.set_marker(colormap[orig_color]['marker'])
                    line.set_markersize(MARKERSIZE)


def set_ax_legend_to_bw(ax, style, colormap, line_style='continuous'):
    '''convert axes legend to black and white
    
    Parameters
    ----------
    ax : axes
    style : {GREYSCALE, HATCHING}
    colormap : dict
               mapping of color to B&W rendering
    line_style: str
                linestyle to use for converting, can be continuous, black
                or None
    
    '''
    
    legend = ax.legend_
    set_legend_to_bw(legend, style, colormap, line_style)


def set_fig_to_bw(fig, style=HATCHING, line_style='continuous', all_colors=None):
    """
    TODO it would be nice if for lines you can select either markers, gray 
    scale, or simple black
    
    Take each axes in the figure and transform its content to black and white. 
    Lines are tranformed based on different line styles. Fills such as can 
    be used in `meth`:envelopes are gray-scaled. Heathmaps are also gray-scaled.
    
    
    derived from and expanded for my use from:
    http://stackoverflow.com/questions/7358118/matplotlib-black-white-colormap-with-dashes-dots-etc
    
    Parameters
    ----------
    fig : figure
          the figure to transform to black and white
    style : {HATCHING, GREYSCALE}
            parameter controlling how collections are transformed.
    line_style: str
                linestyle to use for converting, can be continuous, black
                or None  
    
    """
    all_colors = _identify_colors(fig)
    
    if len(all_colors)>len(bw_mapping):
        mapping_cycle = itertools.cycle(bw_mapping)
        ema_logging.warning('more colors used than provided in B&W mapping, cycling over mapping')
    else:
        mapping_cycle = bw_mapping 
    colormap = dict(zip(all_colors, mapping_cycle))
    ema_logging.debug(colormap.keys())
    
    max_shade=0.9
    for i, color in enumerate(colormap.keys()):
        relative_color=max_shade*((i+1)/len(all_colors))
        colormap[color]['fill'] = str(relative_color)
    
    for ax in fig.get_axes():
        set_ax_lines_bw(ax, colormap)
        set_ax_patches_bw(ax, colormap)
        set_ax_collections_to_bw(ax, style, colormap)
        set_ax_legend_to_bw(ax, style, colormap)
        
    set_legend_to_bw(fig.legends, style, colormap)
    



'''
Created on May 26, 2015

@author: jhkwakkel
'''
import numpy as np
import matplotlib.pyplot as plt

import ema_workbench.analysis.cart as cart
from ema_workbench import ema_logging, load_results

ema_logging.log_to_stderr(level=ema_logging.INFO)

def classify(data):
    #get the output for deceased population
    result = data['deceased population region 1']
    
    #make an empty array of length equal to number of cases 
    classes =  np.zeros(result.shape[0])
    
    #if deceased population is higher then 1.000.000 people, classify as 1 
    classes[result[:, -1] > 1000000] = 1
    
    return classes

#load data
fn = r'./data/1000 flu cases.tar.gz'
results = load_results(fn)
experiments, results = results

#extract results for 1 policy
logical = experiments['policy'] == 'no policy'
new_experiments = experiments[ logical ]
new_results = {}
for key, value in results.items():
    new_results[key] = value[logical]

results = (new_experiments, new_results)

#perform cart on modified results tuple

cart_alg = cart.setup_cart(results, classify, mass_min=0.05)
cart_alg.build_tree()

#print cart to std_out
print cart_alg.stats_to_dataframe()
print cart_alg.boxes_to_dataframe()

#visualize
cart_alg.display_boxes(together=True)
plt.show()


import warnings

warnings.simplefilter("once", ImportWarning)

try:
    from . import vensim
except ImportError:
    warnings.warn("vensim connector not available", ImportWarning)

try:
    from . import excel
except ImportError:
    warnings.warn("excel connector not available", ImportWarning)

try:
    from . import netlogo
except ImportError:
    warnings.warn("netlogo connector not available", ImportWarning)

try:
    from . import pysd_connector
except ImportError:
    warnings.warn("pysd connector not available", ImportWarning)




'''
This module provides functionality for combining the EMA workbench
with IPython parallel. 

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import collections
import logging
import os
import shutil
import socket
import threading

import zmq
from zmq.eventloop import ioloop, zmqstream

from traitlets.config import Application
from traitlets.config.configurable import LoggingConfigurable
from traitlets import Unicode, Instance, List

from ipyparallel.engine.log import EnginePUBHandler
from jupyter_client.localinterfaces import localhost

from . import experiment_runner
from .ema_multiprocessing import setup_working_directories
from .model import AbstractModel
from .util import NamedObjectMap
from ..util import ema_exceptions, ema_logging

# Created on Jul 16, 2015
# 
# .. codeauthor::  jhkwakkel

SUBTOPIC = ema_logging.LOGGER_NAME
engine = None

class EngingeLoggerAdapter(logging.LoggerAdapter):
    '''LoggerAdapter that inserts EMA as a topic into log messages
    
    Parameters
    ----------
    logger : logger instance 
    topic : str
    
    '''

    def __init__(self, logger, topic):
        self.logger = logger
        self.topic = topic
        
    def process(self, msg, kwargs):
        msg = '{topic}::{msg}'.format(topic=self.topic, msg=msg)
        return msg, kwargs

def start_logwatcher():
    '''convenience function for starting the LogWatcher 
    
    Returns
    -------
    LogWatcher
        the log watcher instance
    Thread
        the log watcher thread
        
    .. note : there can only be one log watcher on a given url. 
    
    '''

    logwatcher = LogWatcher()
    
    def starter():
        logwatcher.start()
        try:
            logwatcher.loop.start()
        except (zmq.error.ZMQError, IOError, OSError):
            ema_logging.warning('shutting down log watcher')
        except RuntimeError:
            pass
    
    logwatcher_thread = threading.Thread(target=starter)
    logwatcher_thread.deamon = True
    logwatcher_thread.start()
    
    return logwatcher, logwatcher_thread


def set_engine_logger():
    '''Updates EMA logging on the engines with an EngineLoggerAdapter 
    This adapter injects EMA as a topic into all messages
    '''
    
    logger = Application.instance().log
    logger.setLevel(ema_logging.DEBUG)

    for handler in logger.handlers:
        if isinstance(handler, EnginePUBHandler): # @UndefinedVariable
            handler.setLevel(ema_logging.DEBUG)
    
    adapter = EngingeLoggerAdapter(logger, SUBTOPIC)
    ema_logging._logger = adapter
    
    ema_logging.debug('updated logger')
    

def get_engines_by_host(client):
    ''' returns the engine ids by host
    
    Parameters
    ----------
    client : IPython.parallel.Client instance
    
    Returns
    -------
    dict
        a dict with hostnames as keys, and a list
        of engine ids
    
    '''
    
    results = {i:client[i].apply_sync(socket.gethostname) for i in client.ids}

    engines_by_host = collections.defaultdict(list)
    for engine_id, host in results.items():
        engines_by_host[host].append(engine_id)
    return engines_by_host


def update_cwd_on_all_engines(client):
    ''' updates the current working directory on the engines to point to the 
    same working directory as this notebook
    
    currently only works if engines are on same 
    machine.
    
    Parameters
    ----------
    client : IPython.parallel.Client instance
    
    '''

    engines_by_host = get_engines_by_host(client)
    
    notebook_host = socket.gethostname()
    for key, value in engines_by_host.items():

        if key == notebook_host:
            cwd = os.getcwd()

            # easy, we know the correct cwd
            for engine in value:
                client[engine].apply(os.chdir, cwd)
        else:
            raise NotImplementedError('not yet supported')


class LogWatcher(LoggingConfigurable):
    """A simple class that receives messages on a SUB socket, as published
    by subclasses of `zmq.log.handlers.PUBHandler`, and logs them itself.
    
    This can subscribe to multiple topics, but defaults to all topics.
    """
    
    # configurables
    topics = List([''],
        help="The ZMQ topics to subscribe to. Default is to subscribe to all messages").tag(config=True)
    url = Unicode(
        help="ZMQ url on which to listen for log messages").tag(config=True)
    def _url_default(self):
        return 'tcp://%s:20202' % localhost()
    
    # internals
    stream = Instance('zmq.eventloop.zmqstream.ZMQStream', allow_none=True)
    
    context = Instance(zmq.Context)
    def _context_default(self):
        return zmq.Context.instance()
    
    loop = Instance(zmq.eventloop.ioloop.IOLoop)  # @UndefinedVariable
    def _loop_default(self):
        return ioloop.IOLoop.instance()
    
    def __init__(self, **kwargs):
        super(LogWatcher, self).__init__(**kwargs)
        self.s = self.context.socket(zmq.SUB)  # @UndefinedVariable
        self.s.bind(self.url)
        self.stream = zmqstream.ZMQStream(self.s, self.loop)
        self.subscribe()
        self.observe(self.subscribe, 'topics')
    
    def start(self):
        self.stream.on_recv(self.log_message)
    
    def stop(self):
        self.stream.setsockopt(zmq.UNSUBSCRIBE, b'')  # @UndefinedVariable
        self.stream.stop_on_recv()
        self.s.close()
    
    def subscribe(self):
        """Update our SUB socket's subscriptions."""
        self.stream.setsockopt(zmq.UNSUBSCRIBE, b'')  # @UndefinedVariable
        if '' in self.topics:
            self.log.debug("Subscribing to: everything")
            self.stream.setsockopt(zmq.SUBSCRIBE, b'')  # @UndefinedVariable
        else:
            for topic in self.topics:
                self.log.debug("Subscribing to: %r"%(topic))
                self.stream.setsockopt(zmq.SUBSCRIBE, topic)  # @UndefinedVariable
    
    def _extract_level(self, topic_str):
        """Turn 'engine.0.INFO.extra' into (logging.INFO, 'engine.0.extra')"""
        topics = topic_str.split('.')
        for idx,t in enumerate(topics):
            level = getattr(logging, t, None)
            if level is not None:
                break
        
        if level is None:
            level = logging.INFO
        else:
            topics.pop(idx)
        
        return level, '.'.join(topics)
            
            
    def log_message(self, raw):
        """receive and parse a message, then log it."""
        raw = [r.decode("utf-8") for r in raw]
        
        if len(raw) != 2 or '.' not in raw[0]:
            logging.getLogger().error("Invalid log message: %s"%raw)
            return
        else:
            topic, msg = raw
            level, topic = self._extract_level(topic)
            
            # bit of a hacky way to filter messages
            # assumes subtopic only contains a single dot
            # main topic is now the substring with the origin of the message
            # so e.g. engine.1
            main_topic, subtopic = topic.rsplit('.',1) 
            log = logging.getLogger(subtopic) 
            
            if msg[-1] == '\n':
                msg = msg[:-1]
#             self.log.log(level, "[%s] %s" % (topic, msg))
#             print("[%s] %s" % (main_topic, msg))

            log.log(level, "[%s] %s" % (main_topic, msg))


class Engine(object):
    '''class for setting up ema specific stuff on each engine
    also functions as a convenient namespace for workbench
    relevant variables
    
    Parameters
    ----------
    engine_id : int
    msis : list
    cwd : str
    
    '''

    def __init__(self, engine_id, msis, cwd):
        ema_logging.debug("starting engine {}".format(engine_id))
        self.engine_id = engine_id
        self.msis = msis
        
        ema_logging.debug("setting root working directory to {}".format(cwd))        
        os.chdir(cwd)
        
        models = NamedObjectMap(AbstractModel)
        models.extend(msis)
        self.runner = experiment_runner.ExperimentRunner(models)
        
        self.tmpdir = setup_working_directories(msis, os.getcwd())

    def cleanup_working_directory(self):
        '''remove the root working directory of the engine'''
        if self.tmpdir:
            shutil.rmtree(self.tmpdir) 
            
    def run_experiment(self, experiment):
        '''run the experiment, the actual running is delegated
        to an ExperimentRunner instance'''
        
        try:
            return self.runner.run_experiment(experiment) 
        except ema_exceptions.EMAError:
            raise
        except Exception:
            raise ema_exceptions.EMAParallelError(str(Exception))
       

def initialize_engines(client, msis, cwd):
    '''initialize engine instances on all engines
    
    Parameters
    ----------
    client : IPython.parallel.Client 
    msis : dict
           dict of model structure interfaces with their names as keys
    cwd : str
    
    '''
    for i in client.ids:
        client[i].apply_sync(_initialize_engine, i, msis, cwd)


def cleanup(client):
    '''cleanup directory tree structure on all engines '''
    client[:].apply_sync(_cleanup)


# engines can only deal with functions, not with object.method calls
# these functions are wrappers around the relevant Engine methods
# the engine instance is part of the namespace of the module. 
def _run_experiment(experiment):
    return experiment, engine.run_experiment(experiment)


def _initialize_engine(engine_id, msis, cwd):
    global engine
    engine = Engine(engine_id, msis, cwd)

    
def _cleanup():
    global engine
    engine.cleanup_working_directory()
    del engine



from __future__ import (print_function, absolute_import)

import io
import os
import re

from setuptools import setup


def read(path, encoding='utf-8'):
    path = os.path.join(os.path.dirname(__file__), path)
    with io.open(path, encoding=encoding) as fp:
        return fp.read()


def version(path):
    """Obtain the packge version from a python file e.g. pkg/__init__.py
    See <https://packaging.python.org/en/latest/single_source_version.html>.
    """
    version_file = read(path)
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def package_files(root, package_data_dir):
    paths = []
    
    directory = os.path.join(root, package_data_dir)
    
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            full_path = os.path.join(path, filename)
            
            paths.append(os.path.relpath(full_path, root))
    return paths

example_data_files = package_files('ema_workbench/examples', 'data')
example_data_files = [''.join(['examples/', entry]) for entry in example_data_files]

example_model_files = package_files('ema_workbench/examples', 'models')
example_model_files = [''.join(['examples/', entry]) for entry in example_model_files]

java_files = package_files('ema_workbench/connectors', 'java')
java_files = [''.join(['connectors/', entry]) for entry in java_files]

name = 'ema_workbench'
pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))
pkg_root = pjoin(here, name)

packages = []
for d, _, _ in os.walk(pjoin(here, name)):
    if os.path.exists(pjoin(d, '__init__.py')):
        packages.append(d[len(here)+1:].replace(os.path.sep, '.'))


VERSION = version('ema_workbench/__init__.py')
LONG_DESCRIPTION ="""Project Documentation: https://emaworkbench.readthedocs.io/"""
EXAMPLE_DATA = example_data_files + example_model_files
JAVA = java_files
PACKAGES = packages

setup(
    name            = 'ema_workbench',
    version         = VERSION,
    description     = 'exploratory modelling in Python',
    long_description= LONG_DESCRIPTION,
    author          = 'Jan Kwakkel',
    author_email    = 'j.h.kwakkel@tudelft',
    packages        = PACKAGES,
    package_data    = {'ema_workbench': EXAMPLE_DATA+JAVA},
    url             = 'https://github.com/quaquel/EMAworkbench',
    license         = 'BSD 3-Clause',
    platforms       = "Linux, Mac OS X, Windows",
)



'''
Created on 27 Jul. 2011

This file illustrated the use the EMA classes for a model in Excel.

It used the excel file provided by 
`A. Sharov <http://home.comcast.net/~sharov/PopEcol/lec10/fullmod.html>`_

This excel file implements a simple predator prey model.

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
'''
from __future__ import (division, print_function, absolute_import, 
                        unicode_literals)

from ema_workbench import (RealParameter, TimeSeriesOutcome, ema_logging,
                           perform_experiments)

from ema_workbench.connectors.excel import ExcelModel
from ema_workbench.em_framework.evaluators import MultiprocessingEvaluator


if __name__ == "__main__":    
    ema_logging.log_to_stderr(level=ema_logging.DEBUG)
    
    model = ExcelModel("predatorPrey", wd="./models/excelModel",
                      model_file='/excel example.xlsx')
    model.uncertainties = [RealParameter("K2", 0.01, 0.2), #we can refer to a cell in the normal way
                           RealParameter("KKK", 450,550), # we can also use named cells
                           RealParameter("rP", 0.05,0.15),
                           RealParameter("aaa", 0.00001,0.25),
                           RealParameter("tH", 0.45,0.55),
                           RealParameter("kk", 0.1,0.3)]
    
    #specification of the outcomes
    model.outcomes = [TimeSeriesOutcome("B4:B1076"),  #we can refer to a range in the normal way
                      TimeSeriesOutcome("P_t")] # we can also use named range
    
    #name of the sheet
    model.sheet = "Sheet1"
    
    with MultiprocessingEvaluator(model) as evaluator:
        results = perform_experiments(model, 100, reporting_interval=1,
                                      evaluator=evaluator)
    
    print("blaat")




'''


'''
from __future__ import (unicode_literals, print_function, absolute_import,
                                        division)

import multiprocessing
import numbers 
import os
import random
import shutil
import string
import threading
import warnings

warnings.simplefilter("once", ImportWarning)

from .callbacks import DefaultCallback
from .ema_multiprocessing import LogQueueReader, initializer, add_tasks
from .ema_ipyparallel import (start_logwatcher, set_engine_logger, 
                              initialize_engines, cleanup, _run_experiment)
from .experiment_runner import ExperimentRunner
from .model import AbstractModel
from .optimization import (evaluate_robust, evaluate, EpsNSGAII, 
                           to_dataframe, to_problem, to_robust_problem,
                           process_levers, process_uncertainties, 
                           process_robust)
from .outcomes import ScalarOutcome, AbstractOutcome
from .parameters import (experiment_generator, Scenario, Policy)
from .samplers import (MonteCarloSampler, FullFactorialSampler, LHSSampler, 
                       PartialFactorialSampler, sample_levers, 
                       sample_uncertainties)
from .salib_samplers import (SobolSampler, MorrisSampler, FASTSampler) # TODO:: should become optional import
from .util import NamedObjectMap, determine_objects
from ..util import ema_logging, EMAError


# Created on 5 Mar 2017
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

LHS = 'lhs'
MC = 'mc'
FF = 'ff'
PFF = 'pff'
SOBOL = 'sobol'
MORRIS = 'morris'
FAST = 'fast'

#TODO:: better name, samplers lower case conflicts with module name
SAMPLERS = {LHS:LHSSampler,
            MC:MonteCarloSampler,
            FF:FullFactorialSampler,
            PFF:PartialFactorialSampler,
            SOBOL:SobolSampler,
            MORRIS:MorrisSampler,
            FAST:FASTSampler}

__all__ = ['MultiprocessingEvaluator', 'IpyparallelEvaluator', 
           'optimize', 'perform_experiments', 'SequentialEvaluator']

class BaseEvaluator(object):
    '''evaluator for experiments using a multiprocessing pool
    
    Parameters
    ----------
    msis : collection of models
    searchover : {None, 'levers', 'uncertainties'}, optional
                  to be used in combination with platypus
    union : {None, True, False}, optional
            to be used in combination with platypus, indicates whether
            you want to optimize over the union or the intersection of
            search_over
    
    Raises
    ------
    ValueError
    
    '''
    
    def __init__(self, msis, union=None):
        super(BaseEvaluator, self).__init__()
        
        if isinstance(msis, AbstractModel):
            msis = [msis]
        
        self._msis = msis
        
#         if searchover:
#             if searchover not in {'levers', 'uncertainties'}:
#                 raise ValueError(("search_over must be one of 'levers'"
#                               "or 'uncertainties' not {}".format(searchover)))
#             
#             self.searchover = searchover
#             
#             self.parameters = determine_objects(msis, searchover, union=union)
#             self.parameter_names = [p.name for p in self.parameters]
#             
#             outcomes = determine_objects(msis, "outcomes", union=union)
#             self.outcomes = [o for o in outcomes if
#                              o.kind != AbstractOutcome.INFO]
#             self.outcome_names = [o.name for o in self.outcomes]

    def __enter__(self):
        
        self.initialize()
        
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        
        self.finalize()
        
        if exc_type is not None:
            return False
        
    def initialize(self):
        ''' initialize the evaluator'''
        
        raise NotImplementedError
    
    def finalize(self):
        ''' finalize the evaluator'''
        
        raise NotImplementedError
        
    def evaluate_experiments(self, scenarios, policies, callback):
        '''used by ema_workbench'''
        
        raise NotImplementedError

    
    def evaluate_all(self, jobs, **kwargs):
        '''makes ema_workbench evaluators compatible with Platypus evaluators
        as used by platypus algorithms'''
        
        problem = jobs[0].solution.problem
        searchover = problem.searchover
        
        if searchover == 'levers':
            scenarios, policies = process_levers(jobs)
            jobs_collection = zip(policies, jobs)
        elif searchover == 'uncertainties':
            scenarios, policies = process_uncertainties(jobs)
            jobs_collection = zip(scenarios, jobs)
        elif searchover == 'robust':
            scenarios, policies = process_robust(jobs)
            jobs_collection = zip(policies, jobs)
        else:
            raise NotImplementedError()
        
        experiments, outcomes = perform_experiments(self._msis, 
                                        scenarios=scenarios, policies=policies, 
                                        evaluator=self)


        if searchover in ('levers', 'uncertainties'):
            evaluate(jobs_collection, experiments, outcomes, 
                      problem)
        else:
            evaluate_robust(jobs_collection, experiments, outcomes, 
                             problem)
            
        return jobs

    def perform_experiments(self, scenarios=0, policies=0, evaluator=None, 
                        reporting_interval=None, uncertainty_union=False, 
                        lever_union=False, outcome_union=False, 
                        uncertainty_sampling=LHS, levers_sampling=LHS):
        '''convenience method for performing experiments.
        
        is forwarded to :func:perform_experiments, with evaluator and models
        arguments added in.
        
        '''
        
        return perform_experiments(self._msis, scenarios=scenarios, 
                    policies=policies, evaluator=self, 
                    reporting_interval=reporting_interval, 
                    uncertainty_union=uncertainty_union, lever_union=lever_union, 
                    outcome_union=outcome_union, 
                    uncertainty_sampling=uncertainty_sampling, 
                    levers_sampling=levers_sampling)


    def optimize(self, algorithm=EpsNSGAII, nfe=10000, searchover='levers',
                 **kwargs):
        '''convenience method for outcome optimization.
        
        is forwarded to :func:optimize, with evaluator and models
        arguments added in.
        
        '''
        
        return optimize(self._msis, algorithm=algorithm, nfe=nfe, 
                        searchover=searchover, evaluator=self, **kwargs)

        
    def robust_optimize(self, robustness_functions, scenarios, 
                        algorithm=EpsNSGAII, nfe=10000, searchover='levers',
                        **kwargs):
        '''convenience method for robust optimization.
        
        is forwarded to :func:robust_optimize, with evaluator and models
        arguments added in.
        
        '''
        return robust_optimize(self._msis, robustness_functions, scenarios,
                               self, algorithm=algorithm, nfe=nfe, **kwargs)


class SequentialEvaluator(BaseEvaluator):
    def __init__(self, models, **kwargs):
        super(SequentialEvaluator, self).__init__(models, **kwargs)
    
    def initialize(self):
        pass
    
    def finalize(self):
        pass
    
    def evaluate_experiments(self, scenarios, policies, callback):
        ema_logging.info("performing experiments sequentially")
        
        ex_gen = experiment_generator(scenarios, self._msis, policies)
        
        models = NamedObjectMap(AbstractModel)
        models.extend(self._msis)
        
        cwd = os.getcwd() 
        runner = ExperimentRunner(models)
        for experiment in ex_gen:
            result = runner.run_experiment(experiment)
            callback(experiment, result)
        runner.cleanup()
        os.chdir(cwd)
    

class MultiprocessingEvaluator(BaseEvaluator):
    '''evaluator for experiments using a multiprocessing pool
    
    Parameters
    ----------
    msis : collection of models
    n_processes : int (optional)
    
    
    '''
    
    def __init__(self, msis, n_processes=None, **kwargs):
        super(MultiprocessingEvaluator, self).__init__(msis, **kwargs)
        
        self._pool = None
        self.n_processes = n_processes

    def initialize(self):
        log_queue = multiprocessing.Queue()
    
        log_queue_reader = LogQueueReader(log_queue)
        log_queue_reader.start()
    
        try:
            loglevel = ema_logging._logger.getEffectiveLevel()
        except AttributeError:
            loglevel=30
            
            
        random_part = [random.choice(string.ascii_letters + string.digits) 
                     for _ in range(5)]
        random_part = ''.join(random_part)
        self.root_dir = os.path.abspath("tmp"+random_part)
        os.makedirs(self.root_dir)
    
        self._pool = multiprocessing.Pool(self.n_processes , initializer, 
                                  (self._msis, log_queue, loglevel, self.root_dir))
        ema_logging.info("pool started")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        ema_logging.info("terminating pool")
        
        if exc_type is not None:
            # When an exception is thrown stop accepting new jobs
            # and abort pending jobs without waiting.
            self._pool.terminate()
            return False
        
        super(MultiprocessingEvaluator, self).__exit__(exc_type, exc_value, traceback)

    def finalize(self):
        # Stop accepting new jobs and wait for pending jobs to finish.
        self._pool.close()
        self._pool.join()
        
        shutil.rmtree(self.root_dir)
        
    def evaluate_experiments(self, scenarios, policies, callback):
        ex_gen = experiment_generator(scenarios, self._msis, policies)
        
        add_tasks(self._pool, ex_gen, callback)


class IpyparallelEvaluator(BaseEvaluator):
    '''evaluator for using an ipypparallel pool'''
    

    def __init__(self,  msis, client, **kwargs):
        super(IpyparallelEvaluator, self).__init__(msis, **kwargs)
        self.client = client
        
    def initialize(self):
        import ipyparallel
        
        ema_logging.debug("starting ipyparallel pool")

        try:
            TIMEOUT_MAX = threading.TIMEOUT_MAX
        except AttributeError:
            TIMEOUT_MAX = 1e10  # noqa        
        ipyparallel.client.asyncresult._FOREVER = TIMEOUT_MAX
        # update loggers on all engines
        self.client[:].apply_sync(set_engine_logger)
        
        ema_logging.debug("initializing engines")
        initialize_engines(self.client, self._msis, 
                                                os.getcwd())
        
        self.logwatcher, self.logwatcher_thread = start_logwatcher()
        
        ema_logging.debug("successfully started ipyparallel pool")
        
        ema_logging.info("performing experiments using ipyparallel")
        
        return self


    def finalize(self):
        self.logwatcher.stop()
        cleanup(self.client)
        
        
    def evaluate_experiments(self, scenarios, policies, callback):
        ex_gen = experiment_generator(scenarios, self._msis, policies)
        
        lb_view = self.client.load_balanced_view()
        
        results = lb_view.map(_run_experiment, 
                              ex_gen, ordered=False, block=False)

        for entry in results:
            callback(*entry)
        


def perform_experiments(models, scenarios=0, policies=0, evaluator=None, 
                        reporting_interval=None, uncertainty_union=False, 
                        lever_union=False, outcome_union=False, 
                        uncertainty_sampling=LHS, levers_sampling=LHS):
    '''sample uncertainties and levers, and perform the resulting experiments
    on each of the models
    
    Parameters
    ----------
    models : one or more AbstractModel instances
    scenarios : int or collection of Scenario instances, optional
    policies :  int or collection of Policy instances, optional
    evaluator : Evaluator instance, optional
    reporting interval : int, optional
    uncertainty_union : boolean, optional
    lever_union : boolean, optional
    uncertainty_sampling : {LHS, MC, FF, PFF, SOBOL, MORRIS, FAST}, optional
    lever_sampling : {LHS, MC, FF, PFF, SOBOL, MORRIS, FAST}, optional
    
    
    '''
    if not scenarios and not policies:
        raise EMAError(('no experiments possible since both ' 
                        'scenarios and policies are 0'))
    
    if not scenarios:
        scenarios = [Scenario("None", **{})]
        uncertainties = []
        n_scenarios = 1
    elif(isinstance(scenarios, numbers.Integral)):
        scenarios = sample_uncertainties(models, scenarios, 
             union=uncertainty_union, sampler=SAMPLERS[uncertainty_sampling]())
        uncertainties = scenarios.parameters
        n_scenarios = scenarios.n
    else:
        try:
            uncertainties = scenarios.parameters
            n_scenarios = scenarios.n
        except AttributeError:
            uncertainties = determine_objects(models, "uncertainties", 
                                              union=True)
            if isinstance(scenarios, Scenario):
                scenarios = [scenarios]
            
            uncertainties = [u for u in uncertainties if u.name in 
                             scenarios[0]]
            n_scenarios = len(scenarios)
        
    
    if not policies:
        policies = [Policy("None", **{})]
        levers = []
        n_policies = 1
    elif(isinstance(policies, numbers.Integral)):    
        policies = sample_levers(models, policies, union=lever_union, 
                                 sampler=SAMPLERS[levers_sampling]())
        levers = policies.parameters
        n_policies = policies.n
    else:
        try:
            levers = policies.parameters
            n_policies = policies.n
        except AttributeError:
            levers = determine_objects(models, "levers", union=True)
            if isinstance(policies, Policy):
                policies = [policies]
            
            levers = [l for l in levers if l.name in policies[0]]
            n_policies = len(policies)
    try:
        n_models = len(models)
    except TypeError:
        n_models = 1

    outcomes = determine_objects(models, 'outcomes', union=outcome_union)
    nr_of_exp = n_models * n_scenarios * n_policies 
    
    ema_logging.info("performing {} scenarios * {} policies * {} model(s) = {} experiments".format(n_scenarios, n_policies, n_models, nr_of_exp))
    
    callback = DefaultCallback(uncertainties,
                               levers,
                               outcomes,
                               nr_of_exp,
                               reporting_interval=reporting_interval)
    
    if not evaluator:
        evaluator = SequentialEvaluator(models)
    
    evaluator.evaluate_experiments(scenarios, policies, callback)
    
    if callback.i != nr_of_exp:
        raise EMAError(('some fatal error has occurred while '
                        'running the experiments, not all runs have ' 
                        'completed. expected {} '.format(nr_of_exp),
                        'got {}'.format(callback.i),
                        '{}'.format(type(callback))))
       
    results = callback.get_results()
    ema_logging.info("experiments finished")
    return results



def optimize(models, algorithm=EpsNSGAII, nfe=10000, 
             searchover='levers', evaluator=None, **kwargs):
    '''optimize the model
    
    Parameters
    ----------
    models : 1 or more Model instances
    algorithm : a valid Platypus optimization algorithm
    nfe : int
    searchover : {'uncertainties', 'levers'}
    kwargs : aditional argumenst to pass on to algorithm
    
    Returns
    -------
    pandas DataFrame
    
    
    Raises
    ------
    EMAError if searchover is not one of 'uncertainties' or 'levers'
    NotImplementedError if len(models) > 1
    
    TODO:: constrains are not yet supported
    TODO:: add an optional single policy/scenario to overwrite the defaults
    
    
    '''
    if searchover not in ('levers', 'uncertainties'):
        raise EMAError(("searchover should be one of 'levers' or"
                        "'uncertainties' not {}".format(searchover)))
        
    try:
        if len(models)==1:
            models = models[0]
        else:
            raise NotImplementedError(("optimization over multiple" 
                                      "models yet supported"))
    except TypeError:
        pass
        
    
    problem = to_problem(models, searchover)
    
    # solve the optimization problem
    if not evaluator:
        evaluator = SequentialEvaluator(models)

    optimizer = algorithm(problem, evaluator=evaluator, **kwargs)
    optimizer.run(nfe)
    results = to_dataframe(optimizer, problem.parameter_names, 
                           problem.outcome_names)

    message = "optimization completed, found {} solutions"
    ema_logging.info(message.format(len(optimizer.algorithm.archive)))

    return results


    

def robust_optimize(model, robustness_functions, scenarios, 
                    evaluator=None, algorithm=EpsNSGAII, nfe=10000, 
                    **kwargs):
    '''
    
    Parameters
    ----------
    model : 
    robustness_functions : collection of ScalarOutcomes
    scenarios : 
    evaluator : Evaluator instance
    algorithm : platypus Algorithm instance
    nfe : int
    kwargs : any additional arguments will be passed on to algorithm
    
    
    Raises
    ------
    AssertionError if robustness_function is not a ScalarOutcome,
    if robustness_funcion.kind is INFO, or 
    if robustness_function.function is None
    
    
    robustness functions are scalar outcomes, kind should be MINIMIZE or
    MAXIMIZE, function is the robustness function you want to use.
    
    
    
    '''
    
    for rf in robustness_functions:
        assert(isinstance(rf, ScalarOutcome))
        assert(rf.kind != AbstractOutcome.INFO)
        assert(rf.function != None)
    
    problem = to_robust_problem(model, scenarios, robustness_functions)

    # solve the optimization problem
    if not evaluator:
        evaluator = SequentialEvaluator(model)
        
    optimizer = algorithm(problem, evaluator=evaluator, **kwargs)
    optimizer.run(nfe)
    
    results = to_dataframe(optimizer, problem.parameter_names, 
                           problem.outcome_names)

    message = "optimization completed, found {} solutions"
    ema_logging.info(message.format(len(optimizer.algorithm.archive)))

    return results
    



'''
Created on 8 mrt. 2011

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                epruyt <e.pruyt (at) tudelft (dot) nl>
'''
from __future__ import (division, unicode_literals, print_function, 
                        absolute_import)
from math import exp

from ema_workbench.em_framework import (RealParameter, CategoricalParameter, 
                                        Outcome, perform_experiments)
from ema_workbench.util import ema_logging
from ema_workbench.connectors.vensim import VensimModel

class ScarcityModel(VensimModel):
    def returnsToScale(self, x, speed, scale):
    
        return (x*1000, scale*1/(1+exp(-1* speed * (x-50))))
     
    def approxLearning(self, x, speed, scale, start):
        x = x-start
        loc = 1 - scale
        a = (x*10000, scale*1/(1+exp(speed * x))+loc)
        return a

    def f(self,x, speed, loc):
        return (x/10, loc*1/(1+exp(speed * x)))

    def priceSubstite(self, x, speed, begin, end):
        scale = 2 * end
        start = begin - scale/2
        
        return (x+2000, scale*1/(1+exp(-1* speed * x)) +start)

    def run_model(self, scenario, policy):
        """Method for running an instantiated model structure """
        kwargs = scenario
        loc = kwargs.pop("lookup shortage loc")
        speed = kwargs.pop("lookup shortage speed")
        lookup = [self.f(x/10, speed, loc) for x in range(0,100)]
        kwargs['shortage price effect lookup'] =  lookup
        
        speed = kwargs.pop("lookup price substitute speed")
        begin = kwargs.pop("lookup price substitute begin")
        end = kwargs.pop("lookup price substitute end")
        lookup = [self.priceSubstite(x, speed, begin, end) for x in range(0,100, 10)]
        kwargs['relative price substitute lookup'] = lookup
                        
        scale = kwargs.pop("lookup returns to scale speed")
        speed = kwargs.pop("lookup returns to scale scale")
        lookup = [self.returnsToScale(x, speed, scale) for x in range(0, 101, 10)]
        kwargs['returns to scale lookup'] = lookup
        
        scale = kwargs.pop("lookup approximated learning speed")
        speed = kwargs.pop("lookup approximated learning scale")
        start = kwargs.pop("lookup approximated learning start")
        lookup = [self.approxLearning(x, speed, scale, start) for x in range(0, 101, 10)]
        kwargs['approximated learning effect lookup'] = lookup
        
        super(ScarcityModel, self).run_model(kwargs, policy)


if __name__ == "__main__":
    ema_logging.log_to_stderr(ema_logging.DEBUG)
    
    model = ScarcityModel("scarcity", wd=r'./models/scarcity', 
                          model_file=r'\MetalsEMA.vpm')
        
    model.outcomes = [Outcome('relative market price', time=True),
                      Outcome('supply demand ratio', time=True),
                      Outcome('real annual demand', time=True),
                      Outcome('produced of intrinsically demanded', time=True),
                      Outcome('supply', time=True),
                      Outcome('Installed Recycling Capacity', time=True),
                      Outcome('Installed Extraction Capacity', time=True)]
                        
    model.uncertainties = [
             RealParameter("price elasticity of demand", 0, 0.5),
             RealParameter("fraction of maximum extraction capacity used", 
                             0.6, 1.2),
             RealParameter("initial average recycling cost", 1,4),
             RealParameter("exogenously planned extraction capacity",
                             0, 15000),
             RealParameter("absolute recycling loss fraction", 0.1, 0.5),
             RealParameter("normal profit margin", 0, 0.4),
             RealParameter("initial annual supply", 100000, 120000),
             RealParameter("initial in goods", 1500000, 2500000),
                  
             RealParameter("average construction time extraction capacity",
                             1,  10),
             RealParameter("average lifetime extraction capacity", 20,  40),
             RealParameter("average lifetime recycling capacity", 20, 40),
             RealParameter("initial extraction capacity under construction",
                             5000,  20000),
             RealParameter("initial recycling capacity under construction",
                             5000, 20000),
             RealParameter("initial recycling infrastructure", 5000, 20000),
                  
             #order of delay
             CategoricalParameter("order in goods delay", (1,4,10, 1000)),
             CategoricalParameter("order recycling capacity delay", (1,4,10)),
             CategoricalParameter("order extraction capacity delay", (1,4,10)),
                  
             #uncertainties associated with lookups
             RealParameter("lookup shortage loc", 20, 50),
             RealParameter("lookup shortage speed", 1, 5),
          
             RealParameter("lookup price substitute speed", 0.1, 0.5),
             RealParameter("lookup price substitute begin", 3, 7),
             RealParameter("lookup price substitute end", 15, 25),
                  
             RealParameter("lookup returns to scale speed", 0.01, 0.2),
             RealParameter("lookup returns to scale scale", 0.3, 0.7),
          
             RealParameter("lookup approximated learning speed", 0.01, 0.2),
             RealParameter("lookup approximated learning scale", 0.3, 0.6),
             RealParameter("lookup approximated learning start", 30, 60)]
    

    results = perform_experiments(model, 50)



'''
Created on 21 mrt. 2013

@author: j.h. kwakkel

To do: check Mac support and handling of custom directories

'''
from __future__ import unicode_literals, absolute_import
import os
import re
import sys


import numpy as np
import pandas as pd
import jpype

from ..util.ema_logging import debug, info, warning

__all__ = ['NetLogoException',
           'NetLogoLink']
    
PYNETLOGO_HOME = os.path.dirname(os.path.abspath(__file__))

#Jar supports NetLogo 5.2 or 6.0
module_name = {'5':'netlogoLink_v52.NetLogoLink',
               '6':'netlogoLink_v6.NetLogoLink'}
jar_sep ={'win32':';',
          'darwin':':'}


def find_netlogo(path):
    '''find the most recent version of netlogo in the specified directory
    
    Parameters
    ----------
    path : str
    
    Returns
    -------
    str
    
    Raises
    ------
    IndexError if no netlogo version is found in path
    
    '''
    
    path = os.path.abspath(path)
    netlogo_versions = [entry for entry in os.listdir(path) if 'netlogo' in 
                        entry.lower()]
    
    # sort handles version numbers properly as long as pattern of 
    # directory name is not changed
    netlogo_versions.sort(reverse=True)
    
    return netlogo_versions[0]


def find_jars(path):
    '''find all jars in directory and return as list
    '''
    
    jars = []
    for root, _, files in os.walk(path):
        for file in files:  # @ReservedAssignment
            if file.endswith(".jar"):
                jars.append(os.path.join(root, file))
    return jars


def establish_netlogoversion(path):
    # TODO: python3 compliance
    pattern = re.compile(r'(?:(\d+)\.)?(?:(\d+)\.)?(\*|\d+)$')
    
    netlogo = os.path.basename(os.path.normpath(path))
    match = pattern.search(netlogo)
    version = match.group()
    
    debug('netlogo version is: '+version)
    
    main_version = version[0]
    return main_version


def find_netlogo_windows():
    netlogo = None
    if os.environ['PROGRAMW6432']:
        paths = [os.environ['PROGRAMFILES(X86)'],
                 os.environ['PROGRAMW6432']] 
        for path in paths:
            try:
                netlogo = find_netlogo(path)
            except IndexError:
                pass
            else:
                netlogo = os.path.join(path, netlogo)
    else:
        path = os.environ['PROGRAMFILES']
        try:
            netlogo = find_netlogo(path)
        except IndexError:
            pass
        else:
            netlogo = os.path.join(path, netlogo)
    
    return netlogo


def find_netlogo_mac():
    paths = ['/Applications', 
             os.path.join(os.getenv('HOME'), 'Applications')]
    netlogo = None
    for path in paths:
        try:
            netlogo = find_netlogo(path)
        except IndexError:
            pass
        else:
            netlogo = os.path.join(path, netlogo)
        
    return netlogo


def find_netlogo_linux():
    raise NotImplementedError

def get_netlogo_home():
    if sys.platform=='win32':
        netlogo_home = find_netlogo_windows()
    elif sys.platform=='darwin':
        netlogo_home = find_netlogo_mac()
    else:
        netlogo_home = find_netlogo_linux()
        
    debug('NETLOGO_HOME: '+netlogo_home)
        
    return netlogo_home
    

class NetLogoException(Exception):
    """Basic project specific exception """
    
    pass

    
class NetLogoLink():
    '''
    
    Create a link with NetLogo. Underneath, the NetLogo JVM is started
    through Jpype.
    
    
    :param gui: boolean, if true run NetLogo with gui, otherwise run in 
                headless mode. Defaults to false.
    :param thd: boolean, if true start NetLogo in 3d mode. Defaults to 
                false
    :param nl_dir: string - full path to NetLogo .exe directory, use if
                NetLogo is not in the default location
    :param nl_version: string - NetLogo version under nl_dir (5.2 or 6.0)
    
    '''
    
    def __init__(self, gui=False, thd=False, netlogo_home=None, 
                 netlogo_version=None, jvm_home=None):
        '''
        
        Create a link with NetLogo. Underneath, the NetLogo JVM is started
        through Jpype.
        
        
        :param gui: boolean, if true run NetLogo with gui, otherwise run in 
                    headless mode. Defaults to false.
        :param thd: boolean, if true start NetLogo in 3d mode. Defaults to 
                    false
        :param nl_dir: string - full path to NetLogo .exe directory, use if
                    NetLogo is not in the default location
        :param nl_version: string - NetLogo version under nl_dir (5.2 or 6.0)
        
        '''
        if not netlogo_home:
            netlogo_home = get_netlogo_home()
        if not netlogo_version:
            netlogo_version = establish_netlogoversion(netlogo_home)
        if not jvm_home:
            jvm_home = jpype.get_default_jvm_path()  
        
        
        if not jpype.isJVMStarted():
            jars = find_jars(netlogo_home)
            jars.append(os.path.join(PYNETLOGO_HOME, 'java', 'netlogoLink_combined.jar'))     
            joined_jars = jar_sep[sys.platform].join(jars)
            jarpath = '-Djava.class.path={}'.format(joined_jars)
            try:
                debug("starting jvm: {} {}".format(jvm_home, jarpath))
                jpype.startJVM(jvm_home, jarpath)
                debug("JVM started")  
            except RuntimeError as e:
                warning('failed to start JVM using jvm_home')
                raise e
            #jpype.java.lang.System.setProperty('user.dir', netlogo_home)
            if sys.platform=='darwin':
                debug('set state to headless on mac os')
                jpype.java.lang.System.setProperty("java.awt.headless", "true");            
            
        
        link = jpype.JClass(module_name[netlogo_version])
        debug('NetLogoLink class found')

        if sys.platform == 'darwin' and gui:
            info('on mac only Headless mode is supported')
            gui=False
        
        self.link = link(jpype.java.lang.Boolean(gui),jpype.java.lang.Boolean(thd))
        debug('NetLogoLink class instantiated')
        
            
    def load_model(self, path):
        '''
        
        Load a NetLogo model.
        
        :param path: the absolute path to the NetLogo model
        :raise: IOError in case the  model is not found
        :raise: NetLogoException wrapped around NetLogo exceptions. 
        
        '''
        try:
            self.link.loadModel(path)
        except jpype.JException(jpype.java.io.IOException) as ex:
            raise IOError(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.lang.InterruptedException) as ex:
            raise NetLogoException(ex.message())
        else:
            debug('loaded model successfully')


    def kill_workspace(self):
        '''
        
        Close NetLogo and shut down the JVM.
        
        '''
        
        self.link.killWorkspace()
        
        
    def command(self, netlogo_command):
        '''
        
        Execute the supplied command in NetLogo
        
        :param netlogo_command: a string with a valid netlogo command
        :raises: NetLogoException in case of either a LogoException or 
                CompilerException being raised by netlogo.
        
        '''
        
        try:
            self.link.command(netlogo_command)
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.nvm.EngineException) as ex:
            raise NetLogoException(ex.message())

            
    def report(self, netlogo_reporter):
        '''
        
        Every reporter (commands which return a value) that can be called in 
        the NetLogo Command Center can be called with this method.
        
        :param netlogo_reporter: a valid netlogo reporter 
        :raises: NetlogoException
        
        '''
        
        try:
            result = self.link.report(netlogo_reporter)
            return self._cast_results(result)
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message()) 
        except jpype.JException(jpype.java.lang.Exception) as ex:
            raise NetLogoException(ex.message()) 
        
        
    def patch_report(self, netlogo_reporter):
        '''
        
        Reports a Pandas dataframe with the same shape as the NetLogo world, with
        column labels and row indices following pxcor and pycor patch coordinates.
        The values of the dataframe correspond to patch attributes.
        
        :param netlogo_reporter: valid patch attribute 
        :raises: NetlogoException
        :returns: Dataframe containing NetLogo patch attributes
        
        '''

        try:
            extents = self.link.report('(list min-pxcor max-pxcor min-pycor max-pycor)')
            extents = self._cast_results(extents).astype(int)

            results_df = pd.DataFrame(index=range(extents[2],extents[3]+1,1),
                                      columns=range(extents[0],extents[1]+1,1))
            results_df.sort_index(ascending=False, inplace=True)
            if self.NL_VERSION == '6.0':
                resultsvec = self.link.report('map [[?1] -> [{0}] of ?1] sort patches'.format(netlogo_reporter))
            else:
                resultsvec = self.link.report('map [[{0}] of ?] sort patches'.format(netlogo_reporter))
            resultsvec = self._cast_results(resultsvec)
            results_df.ix[:,:] = resultsvec.reshape(results_df.shape)
            
            return results_df
            
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message()) 
        except jpype.JException(jpype.java.lang.Exception) as ex:
            raise NetLogoException(ex.message())
            
    
    def patch_set(self, attribute, data):
        '''
        
        Set patch attributes from a Pandas dataframe
        
        :param attribute: valid NetLogo patch attribute
        :param data: Pandas dataframe with same dimensions as NetLogo world
        :raises: NetLogoException
        
        '''
    
        try:
            np.set_printoptions(threshold = np.prod(data.shape))
            datalist = '['+str(data.as_matrix().flatten()).strip('[ ')
            datalist = ' '.join(datalist.split())
            if self.NL_VERSION == '6.0':
                command = '(foreach map [[?1] -> [pxcor] of ?1] sort patches map [[?2] -> [pycor] of ?2] \
                            sort patches {0} [[?1 ?2 ?3 ] -> ask patch ?1 ?2 [set {1} ?3]])'.format(datalist, attribute)
            else:
                command = '(foreach map [[pxcor] of ?] sort patches map [[pycor] of ?] \
                            sort patches {0} [ask patch ?1 ?2 [set {1} ?3]])'.format(datalist, attribute)

            self.link.command(command)
            
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message()) 
        except jpype.JException(jpype.java.lang.Exception) as ex:
            raise NetLogoException(ex.message())
        
        
    def repeat_command(self, netlogo_command, reps):
        '''
        
        Execute the supplied command in NetLogo a given number of times
        
        :param netlogo_command: a string with a valid NetLogo command
        :param reps: number of times to repeat commands
        :raises: NetLogoException in case of either a LogoException or 
                CompilerException being raised by NetLogo.
        
        '''
    
        try:
            commandstr = 'repeat {0} [{1}]'.format(reps, netlogo_command)
            self.link.command(commandstr)
        except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
            raise NetLogoException(ex.message())
        except jpype.JException(jpype.java.org.nlogo.nvm.EngineException) as ex:
            raise NetLogoException(ex.message())
        

    def repeat_report(self, netlogo_reporter, reps):
        '''
        
        Execute a reporter (or multiple reporters if a list of strings is given as input)
        over a number of ticks.
        
        :param netlogo_reporter: valid NetLogo reporters (string or list of strings)
        :param reps: number of ticks for which to return values
        :raises: NetLogoException
        :returns: Dataframe of reported values indexed by ticks, with columns 
         for each reporter.
        
        '''
        
        if isinstance(netlogo_reporter, str):
            cols = [netlogo_reporter]
        elif isinstance(netlogo_reporter, list):
            cols = netlogo_reporter   
        else:
            raise NetLogoException("Unknown datatype")
        
        results_df = pd.DataFrame(columns=cols)
        
        for _ in range(reps):    
            for reporter in results_df.columns:
                try:
                    tick = self._cast_results(self.link.report('ticks'))
                    result = self.link.report(reporter)
                    results_df.loc[tick, reporter] = self._cast_results(result)
                except jpype.JException(jpype.java.org.nlogo.api.LogoException) as ex:
                    raise NetLogoException(ex.message())
                except jpype.JException(jpype.java.org.nlogo.api.CompilerException) as ex:
                    raise NetLogoException(ex.message()) 
                except jpype.JException(jpype.java.lang.Exception) as ex:
                    raise NetLogoException(ex.message()) 
            
            self.link.command('go')
                
        return results_df


    def _cast_results(self, results):
        '''
        
        Convert the results to the proper python data type. The NLResults
        object knows its datatype and has converter methods for each.
        
        :param results; the results from report
        :returns: a correct python version of the results
        
        '''
        
        java_dtype = results.type
        
        if java_dtype == "Boolean":
            results = results.getResultAsBoolean()
            if results == 1:
                return True
            else:
                return False
        elif java_dtype == "String":
            return results.getResultAsString()       
        elif java_dtype == "Integer":
            return results.getResultAsInteger()
        elif java_dtype == "Double":
            return results.getResultAsDouble()
        elif java_dtype == "BoolList":
            results = results.getResultAsBooleanArray()
            
            tr = []
            for entry in results:
                if entry == 1:
                    tr.append(True)
                else:
                    tr.append(False)
            return tr
        elif java_dtype == "StringList":
            return results.getResultAsStringArray()   
        elif java_dtype == "IntegerList":
            return results.getResultAsIntegerArray() 
        elif java_dtype == "DoubleList":
            return np.array(results.getResultAsDoubleArray())
        else:
            raise NetLogoException("Unknown datatype")


'''
Created on 20 May, 2011

This module shows how you can use vensim models directly
instead of coding the model in Python. The underlying case
is the same as used in fluExample

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                epruyt <e.pruyt (at) tudelft (dot) nl>
'''
from __future__ import (division, unicode_literals, print_function, 
                        absolute_import)

from ema_workbench import (RealParameter, TimeSeriesOutcome, ema_logging,
                           perform_experiments)

from ema_workbench.connectors.vensim import VensimModel 
from ema_workbench.em_framework.evaluators import MultiprocessingEvaluator

if __name__ == "__main__":
    ema_logging.log_to_stderr(ema_logging.INFO)

    model = VensimModel("fluCase", wd='./models/flu',
                        model_file = 'FLUvensimV1basecase.vpm')
            
    #outcomes
    model.outcomes = [TimeSeriesOutcome('deceased population region 1'),
                      TimeSeriesOutcome('infected fraction R1')]
    
    #Plain Parametric Uncertainties 
    model.uncertainties = [
        RealParameter('additional seasonal immune population fraction R1', 0, 0.5),
        RealParameter('additional seasonal immune population fraction R2', 0, 0.5),
        RealParameter('fatality ratio region 1', 0.0001, 0.1),
        RealParameter('fatality rate region 2', 0.0001, 0.1),
        RealParameter('initial immune fraction of the population of region 1', 0, 0.5),
        RealParameter('initial immune fraction of the population of region 2', 0, 0.5),
        RealParameter('normal interregional contact rate', 0, 0.9),
        RealParameter('permanent immune population fraction R1', 0, 0.5),
        RealParameter('permanent immune population fraction R2', 0, 0.5),
        RealParameter('recovery time region 1', 0.1, 0.75),
        RealParameter('recovery time region 2', 0.1, 0.75),
        RealParameter('susceptible to immune population delay time region 1', 0.5, 2),
        RealParameter('susceptible to immune population delay time region 2', 0.5, 2),
        RealParameter('root contact rate region 1', 0.01, 5),
        RealParameter('root contact ratio region 2', 0.01, 5),
        RealParameter('infection ratio region 1', 0, 0.15),
        RealParameter('infection rate region 2', 0, 0.15),
        RealParameter('normal contact rate region 1', 10, 100),
        RealParameter('normal contact rate region 2', 10, 200)]

    nr_experiments = 10
    with MultiprocessingEvaluator(model) as evaluator:
        results = perform_experiments(model, nr_experiments, 
                                      evaluator=evaluator)



'''
Created on Oct 1, 2012

This is a simple example of the lookup uncertainty provided for 
use in conjuction with vensim models. This example is largely based on 
`Eker et al. (2014) <http://onlinelibrary.wiley.com/doi/10.1002/sdr.1518/suppinfo>`_

@author: sibeleker
@author: jhkwakkel
'''
from __future__ import (division, unicode_literals, print_function, 
                        absolute_import)

import matplotlib.pyplot as plt

from ema_workbench import TimeSeriesOutcome,perform_experiments, ema_logging

from ema_workbench.connectors.vensim import (LookupUncertainty, 
                                             VensimModel)
from ema_workbench.analysis.plotting import lines
from ema_workbench.analysis.plotting_util import BOXPLOT

class Burnout(VensimModel): 
    model_file = r'\BURNOUT.vpm'
    outcomes = [TimeSeriesOutcome('Accomplishments to Date'),
                TimeSeriesOutcome('Energy Level'),
                TimeSeriesOutcome('Hours Worked Per Week'),
                TimeSeriesOutcome('accomplishments per hour')]
    
    def __init__(self, working_directory, name):
        super(Burnout, self).__init__(working_directory, name)
        
        self.uncertainties = [LookupUncertainty('hearne2',[(-1, 3), (-2, 1), (0, 0.9), (0.1, 1), (0.99, 1.01), (0.99, 1.01)], 
                                        "accomplishments per hour lookup", self, 0, 1),
                         LookupUncertainty('hearne2', [(-0.75, 0.75), (-0.75, 0.75), (0, 1.5), (0.1, 1.6), (-0.3, 1.5), (0.25, 2.5)], 
                                         "fractional change in expectations from perceived adequacy lookup", self, -1, 1),
                         LookupUncertainty('hearne2', [(-2, 2), (-1, 2), (0, 1.5), (0.1, 1.6), (0.5, 2), (0.5, 2)], 
                                        "effect of perceived adequacy on energy drain lookup", self, 0, 10),
                         LookupUncertainty('hearne2', [(-2, 2), (-1, 2), (0, 1.5), (0.1, 1.6), (0.5, 1.5), (0.1, 2)], 
                                        "effect of perceived adequacy of hours worked lookup", self, 0, 2.5),
                         LookupUncertainty('hearne2', [(-1, 1), (-1, 1), (0, 0.9), (0.1, 1), (0.5, 1.5), (1, 1.5)], 
                                        "effect of energy levels on hours worked lookup", self, 0, 1.5),
                         LookupUncertainty('hearne2', [(-1, 1), (-1, 1), (0, 0.9), (0.1, 1), (0.5, 1.5), (1, 1.5)], 
                                        "effect of high energy on further recovery lookup", self, 0, 1.25),
                         LookupUncertainty('hearne2', [(-2, 2), (-1, 1), (0, 100), (20, 120), (0.5, 1.5), (0.5, 2)], 
                                        "effect of hours worked on energy recovery lookup", self, 0, 1.5),
                         LookupUncertainty('approximation', [(-0.5, 0.35), (3, 5), (1, 10), (0.2, 0.4), (0, 120)],
                                        "effect of hours worked on energy drain lookup", self, 0, 3),
                         LookupUncertainty('hearne1', [(0, 1), (0, 0.15), (1, 1.5), (0.75, 1.25)], 
                                        "effect of low energy on further depletion lookup", self, 0, 1)]     

        self._delete_lookup_uncertainties()                   

        
if __name__ == "__main__":
    ema_logging.log_to_stderr(ema_logging.INFO)
    model = Burnout(r'./models/burnout', "burnout")
    
    #run policy with old cases
    results = perform_experiments(model, 100)
    lines(results, 'Energy Level', density=BOXPLOT)
    plt.show()

    
